#! /usr/bin/env python
"""Execute events along a timeline.

Examples
--------
>>> timeline = Timeline()
>>> timeline.add_recurring_event('event 1', 1.)
>>> timeline.add_recurring_event('event 2', .3)
>>> timeline.pop()
'event 2'
>>> timeline.pop()
'event 2'
>>> timeline.pop()
'event 2'
>>> timeline.pop()
'event 1'

>>> timeline.time
1.0

>>> for event in timeline.iter_until(2.0): print(event)
event 2
event 2
event 2
event 1

>>> timeline.pop()
'event 2'

When events occur at the same time, events are popped in as first-in, first-out.

>>> timeline = Timeline([('event 1', 1.), ('event 2', .5)])
>>> for event in timeline.iter_until(1.0): print(event)
event 2
event 1
event 2

>>> timeline = Timeline([('event 2', .5), ('event 1', 1.)])
>>> timeline.pop_until(1.05)
['event 2', 'event 1', 'event 2']

>>> timeline.time
1.05

>>> for event in timeline.iter_until(1.05): print(event)
>>> timeline.time
1.05
>>> for event in timeline.iter_until(1.06): print(event)
>>> timeline.time
1.06

The event key can be any object, even a `tuple`.

>>> timeline = Timeline([(('event', 2), .5), (('event', 0), 1.)])
>>> for event in timeline.iter_until(1.05): print(event)
('event', 2)
('event', 0)
('event', 2)

Events do not have to be recurring.

>>> timeline = Timeline([(('event', 2), .5), (('event', 0), 1.)])
>>> timeline.add_one_time_event('one-timer', .1)
>>> for event in timeline.iter_until(1.05): print(event)
one-timer
('event', 2)
('event', 0)
('event', 2)
"""
import bisect


class Timeline(object):
    """Create a timeline of events.

    Parameters
    ----------
    events : dict-like
        Events as event-object/repeat interval pairs.
    start : float, optional
        Start time for the timeline.

    Examples
    --------
    Create a timeline with two recurring events. Events can be any old
    object. In this case they are two strings.

    >>> timeline = Timeline([('hello', 1.), ('world', 1.5)], start=3.)
    >>> sorted(timeline.events) # The events are returned as a set.
    ['hello', 'world']

    The first events will not occur at the starting time.

    >>> timeline.time
    3.0
    >>> timeline.next_event
    'hello'
    >>> timeline.time_of_next_event
    4.0

    To advance the timeline forward to the next event, use the `pop` method.

    >>> timeline.pop()
    'hello'
    >>> timeline.time
    4.0
    >>> timeline.pop()
    'world'
    >>> timeline.time
    4.5

    The timeline keeps track of objects, without making copies. The objects
    don't even need to be hashable.

    >>> hello = ['hello', 'world']
    >>> timeline = Timeline([(hello, 1.)])
    >>> event = timeline.pop()
    >>> event is hello
    True
    >>> event.append('!')
    >>> hello
    ['hello', 'world', '!']
    """

    def __init__(self, events=None, start=0.0):
        events = events or {}

        self._time = float(start)

        self._events = []
        self._times = []
        self._intervals = []

        self.add_recurring_events(events)

    @property
    def time(self):
        """Current time along the timeline.

        Examples
        --------
        >>> timeline = Timeline(start=0)
        >>> timeline.time
        0.0
        >>> timeline = Timeline(start=2)
        >>> timeline.time
        2.0
        """
        return self._time

    @property
    def next_event(self):
        """Next event object.

        Return the next event object but don't advance the timeline forward in
        time.

        Examples
        --------
        >>> timeline = Timeline([('an event', 1.)])
        >>> timeline.next_event
        'an event'
        >>> timeline.time
        0.0
        """
        try:
            return self._events[0]
        except IndexError:
            raise IndexError("empty timeline")

    @property
    def time_of_next_event(self):
        """Time when the next event will happen.

        Examples
        --------
        >>> timeline = Timeline([('an event', 1.)])
        >>> timeline.time_of_next_event
        1.0
        """
        try:
            return self._times[0]
        except IndexError:
            raise IndexError("empty timeline")

    @property
    def events(self):
        """All of the event objects in the timeline.

        Examples
        --------
        >>> timeline = Timeline([('an event', 1.), ('another event', 1.)])
        >>> events = timeline.events
        >>> isinstance(events, set)
        True
        >>> sorted(events)
        ['an event', 'another event']
        """
        return set(self._events)

    def add_recurring_events(self, events):
        """Add a series of recurring events to the timeline.

        Adds recurring events to the timeline. *events* is a list where
        each element is tuple that gives the event object followed by the
        event recurrence interval.

        Parameters
        ----------
        events : dict-like
            Events object/interval pairs to add to the timeline.

        Examples
        --------
        >>> timeline = Timeline()
        >>> len(timeline.events)
        0
        >>> timeline.add_recurring_events([('hello', 1.), ('world', 1.)])
        >>> sorted(timeline.events)
        ['hello', 'world']

        Events pop as first-in, first-out.

        >>> timeline.pop()
        'hello'
        >>> timeline.pop()
        'world'

        The same event can be added multiple times.

        >>> timeline = Timeline()
        >>> timeline.add_recurring_events([('hello', 1.), ('world', 1.),
        ...     ('hello', 1.)])
        >>> sorted(timeline.events)
        ['hello', 'world']
        >>> timeline.pop_until(2.)
        ['hello', 'world', 'hello', 'hello', 'world', 'hello']
        """
        try:
            event_items = events.items()
        except AttributeError:
            event_items = events

        for (event, interval) in event_items:
            self.add_recurring_event(event, interval)

    def add_recurring_event(self, event, interval):
        """Add a recurring event to the timeline.

        Adds a single recurring event to the timeline. *event* is the
        event object, and *interval* is recurrence interval.

        Parameters
        ----------
        event : event-like
            Event to add to the timeline.
        interval : float
            Recurrence interval of the event.

        Examples
        --------
        >>> timeline = Timeline()
        >>> timeline.add_recurring_event('say hello', 1.)
        >>> timeline.events == {'say hello'}
        True
        """
        self._insert_event(event, self.time + interval, interval)

    def add_one_time_event(self, event, time):
        """Add an event that will only happen once.

        Parameters
        ----------
        event : event-like
            Event to add to the timeline.
        time : float
            Time for the event to execute.

        Examples
        --------
        >>> timeline = Timeline()
        >>> timeline.add_one_time_event('say hello', 1.)
        >>> timeline.time_of_next_event
        1.0
        >>> timeline.pop()
        'say hello'
        >>> timeline.time_of_next_event  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        IndexError: empty timeline
        """
        self._insert_event(event, time, None)

    def _insert_event(self, event, time, interval):
        index = bisect.bisect_right(self._times, time)

        self._times.insert(index, time)
        self._events.insert(index, event)
        self._intervals.insert(index, interval)

    def pop(self):
        """Pop the next event from the timeline.

        Examples
        --------
        >>> timeline = Timeline(dict(hello=1.))
        >>> timeline.pop()
        'hello'
        """
        try:
            (event, time, interval) = (
                self._events.pop(0),
                self._times.pop(0),
                self._intervals.pop(0),
            )
        except IndexError:
            raise IndexError("pop from empty timeline")

        try:
            self._insert_event(event, time + interval, interval)
        except TypeError:
            pass

        self._time = time

        return event

    def iter_until(self, stop):
        """Iterate the timeline until a given time.

        Iterate the timeline until *stop*, popping events along the way.

        Parameters
        ----------
        stop : float
            Time to iterate until.

        Returns
        -------
        event
            The next event object as the timeline advances to *stop*.
        """
        if stop < self.time:
            raise ValueError("stop time is less than current time")

        while self.time_of_next_event <= stop:
            yield self.pop()
        self._time = stop

    def pop_until(self, stop):
        """Advance the timeline, popping events along the way.

        Advance the timeline to *stop*, popping events along the way. Returns a
        list of the event objects that were popped to advance the timeline to
        *stop*.

        Parameters
        ----------
        stop : float
            Time to iterate until.

        Returns
        -------
        list
            The events popped to get to the stop time.

        Examples
        --------
        >>> timeline = Timeline([('a', 1.), ('b', 1.), ('c', 1.)])
        >>> timeline.pop_until(2.)
        ['a', 'b', 'c', 'a', 'b', 'c']
        """
        events = []
        for event in self.iter_until(stop):
            events.append(event)
        return events



"""Bridge between BMI and a PyMT component."""
import ctypes
import json
import os
from pprint import pformat

import numpy as np
import yaml
import gimli

from deprecated import deprecated

from ..utils import as_cwd
from ..errors import BmiError
from ..units import transform_azimuth_to_math, transform_math_to_azimuth
from .bmi_docstring import bmi_docstring
from .bmi_mapper import GridMapperMixIn
from .bmi_plot import quick_plot
from .bmi_setup import SetupMixIn
from .bmi_timeinterp import BmiTimeInterpolator
from .bmi_ugrid import dataset_from_bmi_grid


UNITS = gimli.units


class DataValues(object):
    def __init__(self, bmi, name):
        self._bmi = bmi
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def units(self):
        return self._bmi.var_units(self.name)

    @property
    def grid(self):
        if self.location == "none":
            return None
        return self._bmi.var_grid(self.name)

    @property
    def size(self):
        return self._bmi.var_size(self.name)

    @property
    def type(self):
        return self._bmi.var_type(self.name)

    @property
    def intent(self):
        return self._bmi.var_intent(self.name)

    @property
    def location(self):
        return self._bmi.var_grid_loc(self.name)

    @property
    def data(self):
        return self.values()

    def values(self, **kwds):
        if "out" in self.intent:
            return self._bmi.get_value(self.name, **kwds)
        else:
            raise ValueError("not an output var")

    def __repr__(self):
        return str(self)

    def __str__(self):
        return """
<DataValues>
{dtype} {name}({location})
Attributes:
    units: {units}
    grid: {grid}
    intent: {intent}
    location: {location}
""".format(
            dtype=self.type,
            name=self.name,
            units=self.units,
            grid=self.grid,
            intent=self.intent,
            location=self.location,
        ).strip()


class _BmiCapV1(object):

    """Add methods for backward compatibility."""

    @staticmethod
    def _call_bmi(func, *args):
        rtn = func(*args)

        try:
            status, val = rtn
        except TypeError:
            status, val = rtn, None

        if status != 0:
            raise BmiError(func.__name__ + pformat(args), status)
        else:
            return val

    @deprecated(reason="use get_grid_number_of_vertices")
    def get_grid_vertex_count(self, grid):
        return _BmiCapV1._call_bmi(self.bmi.get_grid_vertex_count, grid)

    @deprecated(reason="use get_grid_number_of_faces")
    def get_grid_face_count(self, grid):
        return _BmiCapV1._call_bmi(self.bmi.get_grid_face_count, grid)

    @deprecated(reason="use get_grid_face_node_connectivity")
    def get_grid_connectivity(self, grid, out=None):
        if out is None:
            out = np.empty(self.grid_vertex_count(grid), dtype=ctypes.c_int)
        return _BmiCapV1._call_bmi(self.bmi.get_grid_connectivity, grid, out)

    @deprecated(reason="use get_grid_face_node_offset")
    def get_grid_offset(self, grid, out=None):
        if out is None:
            out = np.empty(self.grid_face_count(grid), dtype=ctypes.c_int)
        return _BmiCapV1._call_bmi(self.bmi.get_grid_offset, grid, out)


class _BmiCapV2(object):
    @deprecated(reason="use get_grid_number_of_vertices")
    def get_grid_vertex_count(self, grid):
        return self.grid_vertex_count(grid)

    @deprecated(reason="use get_grid_number_of_faces")
    def get_grid_face_count(self, grid):
        return self.grid_number_of_faces(grid)

    @deprecated(reason="use get_grid_face_node_connectivity")
    def get_grid_connectivity(self, grid, out=None):
        return self.grid_face_node_connectivity(grid, out=out)

    @deprecated(reason="use get_grid_face_node_offset")
    def get_grid_offset(self, grid, out=None):
        return self.grid_face_node_offset(grid, out=out)


class DeprecatedMethods:
    @deprecated(reason="use grid_ndim")
    def get_grid_rank(self, grid):
        return self.get_grid_ndim(grid)

    @deprecated(reason="use grid_ndim")
    def get_grid_ndim(self, grid):
        return self.grid_ndim(grid)
        # return self.bmi.get_grid_rank(grid)

    @deprecated(reason="use grid_dim")
    def get_grid_dim(self, grid, dim):
        return self.grid_dim(grid, dim)
        # return getattr(self, self.NUMBER_OF_ELEMENTS[dim])(grid)

    @deprecated(reason="use grid_node_count")
    def get_grid_size(self, grid):
        return self.grid_node_count(grid)

    @deprecated(reason="use grid_type")
    def get_grid_type(self, grid):
        return self.grid_type(grid)
        # return self.bmi.get_grid_type(grid)

    @deprecated(reason="use grid_shape")
    def get_grid_shape(self, grid, out=None):
        return self.grid_shape(grid, out=out)

    @deprecated(reason="use grid_spacing")
    def get_grid_spacing(self, grid, out=None):
        return self.grid_spacing(grid, out=out)

    @deprecated(reason="use grid_origin")
    def get_grid_origin(self, grid, out=None):
        return self.grid_origin(grid, out=out)

    @deprecated(reason="use grid_node_count")
    def get_grid_number_of_nodes(self, grid):
        return self.bmi.get_grid_size(grid)

    @deprecated(reason="use grid_edge_count")
    def get_grid_number_of_edges(self, grid):
        return self.grid_edge_count(grid)
        # return self.bmi.get_grid_edge_count(grid)

    @deprecated(reason="use grid_vertex_count")
    def get_grid_number_of_vertices(self, grid):
        return self.grid_vertex_count(grid)
        # return self.get_grid_nodes_per_face(grid).sum()

    @deprecated(reason="use grid_face_count")
    def get_grid_number_of_faces(self, grid):
        return self.grid_face_count(grid)
        # return self.bmi.get_grid_face_count(grid)
        # return self.bmi.get_grid_number_of_faces(grid)

    @deprecated(reason="use grid_face_node_connectivity")
    def get_grid_face_node_connectivity(self, grid, out=None):
        return self.grid_face_node_connectivity(grid, out=out)

    @deprecated(reason="use grid_face_nodes")
    def get_grid_face_nodes(self, grid, out=None):
        return self.grid_face_nodes(grid, out=out)

    @deprecated(reason="use grid_face_node_offset")
    def get_grid_face_node_offset(self, grid, out=None):
        return self.grid_face_node_offset(grid, out=out)

    @deprecated(reason="use grid_nodes_per_face")
    def get_grid_nodes_per_face(self, grid, out=None):
        return self.grid_nodes_per_face(grid, out=out)

    @deprecated(reason="use grid_x")
    def get_grid_x(self, grid, out=None):
        return self.grid_x(grid, out=out)

    @deprecated(reason="use grid_y")
    def get_grid_y(self, grid, out=None):
        return self.grid_y(grid, out=out)

    @deprecated(reason="use grid_z")
    def get_grid_z(self, grid, out=None):
        return self.grid_z(grid, out=out)

    @deprecated(reason="use var_intent")
    def get_var_intent(self, name):
        return self.var_intent(name)

    @deprecated(reason="use var_location")
    def get_var_location(self, name):
        return self.var_location(name)

    @deprecated(reason="use var_grid_loc")
    def get_var_grid_loc(self, name):
        return self.var_loc(name)

    @deprecated(reason="use var_grid")
    def get_var_grid(self, name):
        return self.var_grid(name)

    @deprecated(reason="use var_itemsize")
    def get_var_itemsize(self, name):
        return self.var_itemsize(name)

    @deprecated(reason="use var_nbytes")
    def get_var_nbytes(self, name):
        return self.var_nbytes(name)

    @deprecated(reason="use var_type")
    def get_var_type(self, name):
        return self.var_type(name)

    @deprecated(reason="use var_units")
    def get_var_units(self, name):
        return self.var_units(name)


class _BmiCap(DeprecatedMethods):
    def __init__(self):
        self._bmi = self._cls()
        self._initialized = False
        self._grid = dict()
        self._var = dict()
        self._time_units = None
        self._initdir = None
        super(_BmiCap, self).__init__()

    @property
    def bmi(self):
        return self._bmi

    @property
    def name(self):
        return self.get_component_name()

    @property
    def grid(self):
        return self._grid

    @property
    def var(self):
        return self._var

    @property
    def initdir(self):
        return self._initdir

    def _grid_ids(self):
        grids = set()
        for var in set(self.input_var_names + self.output_var_names):
            if self.var_grid(var) is not None:
                grids.add(self.var_grid(var))
        return tuple(grids)

    def get_component_name(self):
        return self.bmi.get_component_name()

    def initialize(self, fname=None, dir="."):  # pylint: disable=redefined-builtin
        """Initialize the model.

        Parameters
        ----------
        fname : str
            Name of initialization file.
        dir : str
            Path to folder in which to run initialization.
        """
        self._initdir = os.path.abspath(dir)
        with as_cwd(self.initdir):
            self.bmi.initialize(fname or "")
            self._initialized = True

        for grid_id in self._grid_ids():
            self._grid[grid_id] = dataset_from_bmi_grid(self, grid_id)

        for name in set(self.output_var_names + self.input_var_names):
            self._var[name] = DataValues(self, name)

    def update(self):
        with as_cwd(self.initdir):
            return self.bmi.update()

    def finalize(self):
        with as_cwd(self.initdir):
            self._initialized = False
            return self.bmi.finalize()

    def set_value(self, name, val):
        val = np.asarray(val).reshape((-1,))
        return self.bmi.set_value(name, val)

    def get_value(self, name, out=None, units=None, angle=None, at=None, method=None):
        if out is None:
            # grid = self.var_grid(name)
            dtype = self.var_type(name)
            if dtype == "":
                raise ValueError("{name} not understood".format(name=name))
            n_items = self.var_nbytes(name) // self.var_itemsize(name)
            # loc = self.var_grid_loc(name)
            out = np.empty(n_items, dtype=dtype)
            # out = np.empty(self.grid_dim(grid, loc), dtype=dtype)
            # out = np.empty(self.grid_dim(grid, loc), dtype=dtype)

        self.bmi.get_value(name, out)

        if name in self._interpolators and at is not None:
            out[:] = self._interpolators[name].interpolate(at)

        # from_units = Units(self.var_units(name))
        # if units is not None:
        #     to_units = Units(units)
        # else:
        #     to_units = from_units

        # if units is not None and from_units != to_units:
        #     Units.conform(out, from_units, to_units, inplace=True)

        if units is not None:
            convert = UNITS.Unit(self.var_units(name)).to(UNITS[units])
            # convert = UnitConverter(self.var_units(name), units)
            convert(out, out=out)

        # if units is not None:
        #     try:
        #         from_units = self.var_units(name)
        #     except AttributeError, NotImplementedError:
        #         pass
        #     else:
        #         Units.conform(out, Units(from_units), Units(units),
        #                       inplace=True)

        if angle not in ("azimuth", "math", None):
            raise ValueError("angle not understood")

        if units is not None:
            if angle == "azimuth" and "azimuth" not in name:
                transform_math_to_azimuth(out, units)
            elif angle == "math" and "azimuth" in name:
                transform_azimuth_to_math(out, units)

        return out

    def get_value_ptr(self, name):
        return self.bmi.get_value_ptr(name)

    def grid_ndim(self, grid):
        return self.bmi.get_grid_rank(grid)

    NUMBER_OF_ELEMENTS = {
        "node": "grid_node_count",
        "edge": "grid_edge_count",
        "face": "grid_face_count",
        "vertex": "grid_vertex_count",
        # "node": "grid_number_of_nodes",
        # "edge": "grid_number_of_edges",
        # "face": "grid_number_of_faces",
        # "vertex": "grid_number_of_vertices",
    }

    def grid_dim(self, grid, dim):
        return getattr(self, self.NUMBER_OF_ELEMENTS[dim])(grid)

    def grid_type(self, grid):
        return self.bmi.get_grid_type(grid)

    def grid_shape(self, grid, out=None):
        if out is None:
            out = np.empty(self.grid_ndim(grid), dtype=ctypes.c_int)
        self.bmi.get_grid_shape(grid, out)
        return out

    def grid_spacing(self, grid, out=None):
        if out is None:
            out = np.empty(self.grid_ndim(grid), dtype=ctypes.c_double)
        self.bmi.get_grid_spacing(grid, out)
        return out

    def grid_origin(self, grid, out=None):
        if out is None:
            out = np.empty(self.grid_ndim(grid), dtype=ctypes.c_double)
        self.bmi.get_grid_origin(grid, out)
        return out

    def grid_face_node_connectivity(self, grid, out=None):
        if out is None:
            out = np.empty(self.grid_vertex_count(grid), dtype=ctypes.c_int)
        self.bmi.get_grid_face_nodes(grid, out)
        return out

    def grid_face_nodes(self, grid, out=None):
        if self.grid_face_count(grid) > 0:
            if out is None:
                out = np.empty(self.grid_vertex_count(grid), dtype=ctypes.c_int)
            self.bmi.get_grid_face_nodes(grid, out)
        return out

    def grid_face_node_offset(self, grid, out=None):
        nodes_per_face = self.grid_nodes_per_face(grid, out=out)
        return np.cumsum(nodes_per_face, out=out)

    def grid_nodes_per_face(self, grid, out=None):
        if self.grid_face_count(grid) > 0:
            if out is None:
                out = np.empty(self.grid_face_count(grid), dtype=ctypes.c_int)
            self.bmi.get_grid_nodes_per_face(grid, out)
        return out

    def grid_x(self, grid, out=None):
        if out is None:
            if self.grid_type(grid) == "rectilinear":
                out = np.empty(self.grid_shape(grid)[1], dtype=float)
            else:
                out = np.empty(self.grid_node_count(grid), dtype=float)
        self.bmi.get_grid_x(grid, out)
        return out

    def grid_y(self, grid, out=None):
        if out is None:
            if self.grid_type(grid) == "rectilinear":
                out = np.empty(self.grid_shape(grid)[0], dtype=float)
            else:
                out = np.empty(self.grid_node_count(grid), dtype=float)
        self.bmi.get_grid_y(grid, out)
        return out

    def grid_z(self, grid, out=None):
        if out is None:
            if self.grid_type(grid) == "rectilinear":
                shape = self.grid_shape(grid)
                try:
                    zdim = shape[2]
                except IndexError:
                    zdim = 1
                out = np.empty(zdim, dtype=float)
            else:
                out = np.empty(self.grid_node_count(grid), dtype=float)

        self.bmi.get_grid_z(grid, out)
        return out

    def grid_node_count(self, grid):
        try:
            self.bmi.get_grid_node_count
        except AttributeError:
            return self.bmi.get_grid_size(grid)
            # return self.bmi.get_grid_number_of_nodes(grid)
        else:
            return self.bmi.get_grid_node_count(grid)

    def grid_edge_count(self, grid):
        try:
            self.bmi.get_grid_edge_count
        except AttributeError:
            return self.bmi.get_grid_number_of_edges(grid)
        else:
            return self.bmi.get_grid_edge_count(grid)

    def grid_face_count(self, grid):
        try:
            self.bmi.get_grid_face_count
        except AttributeError:
            return self.bmi.get_grid_number_of_faces(grid)
        else:
            return self.bmi.get_grid_face_count(grid)

    def grid_vertex_count(self, grid):
        return (
            self.grid_nodes_per_face(grid).sum()
            if self.grid_face_count(grid) > 0
            else 0
        )
        # return self.grid_nodes_per_face(grid).sum()

    @property
    def input_var_names(self):
        return tuple(self.bmi.get_input_var_names())

    def get_input_var_names(self):
        return self.input_var_names

    @property
    def output_var_names(self):
        return tuple(self.bmi.get_output_var_names())

    def get_output_var_names(self):
        return self.output_var_names

    @property
    def time_units(self):
        return self._time_units or self.bmi.get_time_units()
        # return self.get_time_units()

    @time_units.setter
    def time_units(self, new_units):
        self._time_units = new_units
        self._time_converter = UNITS.Unit(self.bmi.get_time_units()).to(
            UNITS[new_units]
        )
        # self._time_converter = UnitConverter(self.bmi.get_time_units(), new_units)

    # def get_time_units(self):
    #     return self.bmi.get_time_units()

    @property
    def time(self):
        return self._conform_time(self.bmi.get_current_time())

    def _conform_time(self, time):
        try:
            self._time_converter
        except AttributeError:
            return time
        else:
            return self._time_converter(time)

        # try:
        #     self._time_converter
        # except AttributeError:
        #     self._time_converter = UnitConverter(self.bmi.get_time_units())
        # return self._time_converter(time, self.time_units)

    @property
    def start_time(self):
        return self._conform_time(self.bmi.get_start_time())

    @property
    def end_time(self):
        return self._conform_time(self.bmi.get_end_time())

    @property
    def time_step(self):
        return self._conform_time(self.bmi.get_time_step())

    # def get_current_time(self, units=None):
    #     time = self.bmi.get_current_time()
    #     return self.time_in(time, units)

    # def get_start_time(self, units=None):
    #     time = self.bmi.get_start_time()
    #     return self.time_in(time, units)

    # def get_end_time(self, units=None):
    #     time = self.bmi.get_end_time()
    #     return self.time_in(time, units)

    # def get_time_step(self, units=None):
    #     time = self.bmi.get_time_step()
    #     return self.time_in(time, units)

    def time_in(self, time, units):
        # if units is None:
        #     units = self.time_units
        #     # return time

        # try:
        #     units_str = self.time_units
        #     # units_str = self.time_units
        # except (AttributeError, NotImplementedError):
        #     pass
        # else:
        #     from_units = Units(units_str)
        #     to_units = Units(units)

        #     if not from_units.equals(to_units):
        #         time = Units.conform(time, from_units, to_units)

        try:
            self.time_units
        except (AttributeError, NotImplementedError):
            return time

        if units is not None:
            convert = UNITS.Unit(self.time_units).to(UNITS[units])
            # convert = UnitConverter(self.time_units, units)
            time = convert(time)

        return time

    def time_from(self, time, units):
        # if units is None:
        #     return time

        # try:
        #     # units_str = self.time_units
        #     units_str = self.time_units
        # except (AttributeError, NotImplementedError):
        #     pass
        # else:
        #     to_units = Units(units_str)
        #     from_units = Units(units)

        #     if not from_units.equals(to_units):
        #         time = Units.conform(time, from_units, to_units)

        try:
            self.time_units
        except (AttributeError, NotImplementedError):
            return time

        if units is not None:
            convert = UNITS.Unit(units).to(UNITS[self.time_units])
            # convert = UnitConverter(units, self.time_units)
            time = convert(time)

        return time

    def var_intent(self, name):
        intent = ""
        if name in self.input_var_names:
            intent += "in"
        if name in self.output_var_names:
            intent += "out"
        return intent

    def var_location(self, name):
        return self.var_grid_loc(name)

    def var_grid_loc(self, name):
        try:
            self.bmi.get_var_location
        except AttributeError:
            return "node"
        else:
            return self.bmi.get_var_location(name)

    def var_grid(self, name):
        if self.var_location(name) == "none":
            return None
        return self.bmi.get_var_grid(name)

    def var_itemsize(self, name):
        return self.bmi.get_var_itemsize(name)

    def var_nbytes(self, name):
        return self.bmi.get_var_nbytes(name)

    def var_type(self, name):
        return self.bmi.get_var_type(name)

    def var_units(self, name):
        units = self.bmi.get_var_units(name)
        if units == "-":
            return ""
        else:
            return units

    def as_dict(self):
        vars_ = {}
        grid_ids = set()
        for var in set(self.input_var_names + self.output_var_names):
            var_desc = {
                # 'name': var,
                "intent": "",
                "units": self.var_units(var),
                "dtype": self.var_type(var),
                "itemsize": self.var_itemsize(var),
                "nbytes": self.var_nbytes(var),
                "grid": self.var_grid(var),
            }
            vars_[var] = var_desc

            if var in self.input_var_names:
                var_desc["intent"] += "in"
            if var in self.output_var_names:
                var_desc["intent"] += "out"
            # vars_.append(var_desc)
            grid_ids.add(var_desc["grid"])
        # vars_.sort(cmp=lambda a, b: cmp(a['name'], b['name']))

        grids = {}
        for grid_id in grid_ids:
            grid_desc = {
                # 'id': grid_id,
                "rank": self.get_grid_ndim(grid_id),
                "size": self.grid_node_count(grid_id),
                "type": self.get_grid_type(grid_id),
            }
            grids[grid_id] = grid_desc
            # grids.append(grid_desc)
        # grids.sort(cmp=lambda a, b: cmp(a['id'], b['id']))

        in_vars = list(self.input_var_names)
        out_vars = list(self.output_var_names)
        in_vars.sort()
        out_vars.sort()

        times = {
            "start": self.start_time,
            "end": self.end_time,
            "current": self.time,
            # 'time_step': self.get_time_step(),
            "units": self.time_units,
        }
        return {
            "name": self.name,
            "input_var_names": in_vars,
            "output_var_names": out_vars,
            "vars": vars_,
            "grids": grids,
            "times": times,
        }

    def as_yaml(self):
        return yaml.dump(self.as_dict(), default_flow_style=False)

    def as_json(self):
        return json.dumps(self.as_dict())

    def quick_plot(self, name, **kwds):
        return quick_plot(self, name, **kwds)

    def __str__(self):
        return yaml.dump(
            {
                "name": self.name,
                "input_var_names": list(self.input_var_names),
                "output_var_names": list(self.output_var_names),
            },
            default_flow_style=False,
        )


class BmiCap(GridMapperMixIn, _BmiCap, BmiTimeInterpolator, SetupMixIn):
    pass


def bmi_factory(cls):
    class BmiWrapper(BmiCap):
        # __doc__ = bmi_docstring(cls.__name__.split('.')[-1])
        __doc__ = bmi_docstring(cls)
        _cls = cls

        def __str__(self):
            return "{0}".format(cls.__name__)

        def __repr__(self):
            return "<{0}()>".format(cls.__name__)

    BmiWrapper.__name__ = cls.__name__
    return BmiWrapper



#! /usr/bin/env python


class PymtError(Exception):

    pass


class BmiError(PymtError):
    def __init__(self, fname, status):
        self._fname = fname
        self._status = status

    def __str__(self):
        return "error calling BMI function: {fname} ({code})".format(
            fname=self._fname, code=self._status
        )


class BadUnitError(PymtError):
    def __init__(self, unit):
        self._unit = unit

    def __str__(self):
        return "unknown unit ({0!r})".format(self._unit)


class IncompatibleUnitsError(PymtError):
    def __init__(self, src, dst):
        self._src = src
        self._dst = dst

    def __str__(self):
        return "incompatible units ({0!r}, {1!r})".format(self._src, self._dst)



"""Access to framework services.
"""
import warnings

_COMPONENT_CLASSES = {}
_COMPONENT_INSTANCES = {}


def register_component_classes(components, **kwds):
    """Register a list of components with the framework.

    Parameters
    ----------
    components : list
        Component names to register.
    if_exists : {'raise', 'warn', 'pass', 'clobber'}, optional
        What to do if the class is already registered.

    See Also
    --------
    :func:`register_component_class` : Register just a single component.
    """
    for name in components:
        register_component_class(name, **kwds)


def register_component_class(name, if_exists="raise"):
    """Register a component with the framework.

    Add the component class, *name*, to the list of framework services. The
    component name should be the fully-qualified name of a Python class.

    Parameters
    ----------
    name : str
        Name of component to register.
    if_exists : {'raise', 'warn', 'pass', 'clobber'}, optional
        What to do if the class is already registered.

    Raises
    ------
    ValueError
        If the class is already registered with the framework.
    ImportError
        If there is a problem importing any part of *name*.

    Notes
    -----
    This function will try to import *name* as though it were a fully-qualified
    name of a Python class. That is, if *name* were `foo.bar.Baz`, try the
    following::

        from foo.bar import Baz

    Examples
    --------
    >>> from pymt.framework.services import (
    ...     del_services,
    ...     register_component_class,
    ...     get_component_class
    ... )

    >>> del_services()
    >>> register_component_class('pymt.testing.services.AirPort')
    >>> get_component_class('AirPort')
    <class 'pymt.testing.services.AirPort'>

    Raise an ImportError if the component class could not be loaded.

    >>> register_component_class('pymt.testing.services.NotAClass') # doctest : +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ImportError: cannot import component NotAClass from pymt.testing.services
    >>> register_component_class('pymt.not.a.module.AirPort') # doctest : +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ValueError: component class exists (AirPort)
    """
    parts = name.split(".")
    module_name, cls_name = (".".join(parts[:-1]), parts[-1])

    if cls_name in _COMPONENT_CLASSES:
        if if_exists == "raise":
            raise ValueError("component class exists (%s)" % cls_name)
        elif if_exists == "warn":
            warnings.warn("component class exists (%s)" % cls_name)
            return
        elif if_exists == "pass":
            return

    mod = __import__(module_name, fromlist=[cls_name])
    try:
        _COMPONENT_CLASSES[cls_name] = getattr(mod, cls_name)
    except (KeyError, AttributeError):
        raise ImportError(
            "cannot import component %s from %s" % (cls_name, module_name)
        )


def register_component_instance(name, instance):
    """Register a component instance with the framework.

    Parameters
    ----------
    name : str
        Name of the component instance.
    instance : object
        Component instance.

    Raises
    ------
    ValueError
        If *name* is already registered with the framework.

    Examples
    --------
    >>> from pymt.framework.services import (
    ...     del_services,
    ...     register_component_class,
    ...     get_component_instance
    ... )

    >>> del_services()
    >>> from pymt.testing.services import AirPort
    >>> air_port = AirPort()
    >>> register_component_instance('air_port', air_port)
    >>> air_port is get_component_instance('air_port')
    True
    """
    if name in _COMPONENT_INSTANCES:
        raise ValueError("component instance exists (%s)" % name)
    else:
        _COMPONENT_INSTANCES[name] = instance


def get_component_class(name):
    """Get a registered component class by name.

    Parameters
    ----------
    name : str
        Name of the registered class.

    Returns
    -------
    class
        The registered class.

    Raises
    ------
    KeyError
        If the *name* is not registered with the framework.

    See Also
    --------
    :func:`register_component_class`
    """
    return _COMPONENT_CLASSES[name]


def instantiate_component(cls_name, instance_name):
    """Instantiate a registered component class.

    Instantiate a registered class and register that instance with the framework
    as *instance_name*.

    Parameters
    ----------
    cls_name : str
        Name of the registered class.
    instance_name : str
        Name of the new instance.

    Returns
    -------
    object
        An instance of the registered class.

    Examples
    --------
    >>> from pymt.framework.services import (
    ...     del_services,
    ...     register_component_class,
    ...     instantiate_component,
    ...     get_component_instance
    ... )

    >>> del_services()
    >>> register_component_class('pymt.testing.services.AirPort')
    >>> air_port = instantiate_component('AirPort', 'air_port')
    >>> air_port is get_component_instance('air_port')
    True
    """
    register_component_instance(instance_name, get_component_class(cls_name)())
    return get_component_instance(instance_name)


def get_component_instance(name):
    """Get a registered instance from the framework.

    Parameters
    ----------
    name : str
        Name of the registered instance.

    Returns
    -------
    object
        The registered instance.

    Raises
    ------
    KeyError
        If *name* is not registered with the framework.

    See Also
    --------
    :func:`instantiate_component`
    """
    return _COMPONENT_INSTANCES[name]


def del_component_instance(name):
    """Remove an instance by name.

    Parameters
    ----------
    name : str
        Component instance.

    See Also
    --------
    :func:`del_component_instances`
    """
    try:
        del _COMPONENT_INSTANCES[name]
    except KeyError:
        pass


def del_component_instances(names):
    """Remove a list of instances by name.

    Parameters
    ----------
    names : list
        Component instance names.

    See Also
    --------
    :func:`del_component_instance`
    """
    for name in names:
        del_component_instance(name)


def del_services():
    del_component_instances(get_component_instance_names())
    for name in list(_COMPONENT_CLASSES.keys()):
        del _COMPONENT_CLASSES[name]


def get_component_instance_names():
    """Names of all instances.

    Returns
    -------
    list
        Names of all the instanciated components.
    """
    return list(_COMPONENT_INSTANCES.keys())



#! /usr/bin/env python
from ..errors import BmiError
from ..utils import as_cwd
from .timeinterp import TimeInterpolator


class BmiTimeInterpolator(object):
    def __init__(self, *args, **kwds):
        method = kwds.pop("method", "linear")
        self._interpolators = dict(
            [(name, None) for name in self.output_var_names if "__" in name]
        )
        self.reset(method=method)

        super(BmiTimeInterpolator, self).__init__(*args, **kwds)

    def reset(self, method="linear"):
        for name in self._interpolators:
            self._interpolators[name] = TimeInterpolator(method=method)

    def add_data(self):
        for name in self._interpolators:
            try:
                self._interpolators[name].add_data([(self.time, self.get_value(name))])
            except BmiError:
                self._interpolators.pop(name)
                print("unable to get value for {name}. ignoring".format(name=name))

    def interpolate(self, name, at):
        return self._interpolators[name].interpolate(at)

    def update_until(self, then, method=None, units=None):
        with as_cwd(self.initdir):
            then = self.time_from(then, units)

            if hasattr(self.bmi, "update_until"):
                try:
                    self.bmi.update_until(then)
                except NotImplementedError:
                    pass

            self.reset()
            while self.time < then:
                if self.time + self.time_step > then:
                    self.add_data()
                self.update()

            if self.time > then:
                self.add_data()



#! /bin/env python

"""
Examples
--------

Create a grid of length 2 in the x direction, and 3 in the y direction.

>>> (x, y) = np.meshgrid ([1., 2., 4., 8.], [1., 2., 3.])
>>> g = Structured(y.flatten(), x.flatten(), [3, 4])
>>> g.get_point_count()
12
>>> g.get_cell_count()
6
>>> g.get_x()
array([ 1.,  2.,  4.,  8.,  1.,  2.,  4.,  8.,  1.,  2.,  4.,  8.])
>>> g.get_y()
array([ 1.,  1.,  1.,  1.,  2.,  2.,  2.,  2.,  3.,  3.,  3.,  3.])
>>> g.get_shape()
array([3, 4])

Create a grid of length 2 in the i direction, and 3 in the j direction.

>>> (x, y) = np.meshgrid ([1., 2., 4., 8.], [1., 2., 3.])
>>> g = Structured (y.flatten (), x.flatten (), (3, 4), indexing='ij')
>>> g.get_x()
array([ 1.,  2.,  4.,  8.,  1.,  2.,  4.,  8.,  1.,  2.,  4.,  8.])
>>> g.get_y()
array([ 1.,  1.,  1.,  1.,  2.,  2.,  2.,  2.,  3.,  3.,  3.,  3.])
>>> g.get_shape()
array([3, 4])

>>> g.get_offset()
array([ 4,  8, 12, 16, 20, 24])
>>> g.get_connectivity()
array([ 0,  1,  5,  4,  1,  2,  6,  5,  2,  3,  7,  6,  4,  5,  9,  8,  5,
        6, 10,  9,  6,  7, 11, 10])

**Structured grid of points**

The same grid as the previous example but without any cells - just points.

>>> (x, y) = np.meshgrid ([1., 2., 4., 8.], [1., 2., 3.])
>>> g = StructuredPoints (y.flatten (), x.flatten (), (3, 4), indexing='ij', set_connectivity=True)
>>> g.get_x()
array([ 1.,  2.,  4.,  8.,  1.,  2.,  4.,  8.,  1.,  2.,  4.,  8.])
>>> g.get_y()
array([ 1.,  1.,  1.,  1.,  2.,  2.,  2.,  2.,  3.,  3.,  3.,  3.])
>>> g.get_shape()
array([3, 4])

The number of points are the same, but the number of cells is now 0.

>>> g.get_point_count()
12
>>> g.get_cell_count()
0

The offset runs from 1 up to (and including) the number of points.

>>> all (g.get_offset ()==np.arange (1, g.get_point_count ()+1))
True

The connectivity runs from 0 to one less than the number of points.

>>> all (g.get_connectivity ()==np.arange (g.get_point_count ()))
True

**1D Grid of line segments**

>>> g = Structured([-1, 2, 3, 6], (4, ))
>>> g.get_x()
array([-1.,  2.,  3.,  6.])
>>> g.get_y()
Traceback (most recent call last):
        ...
IndexError: Dimension out of bounds
>>> g.get_shape()
array([4])

>>> g.get_offset()
array([2, 4, 6])
>>> g.get_connectivity()
array([0, 1, 1, 2, 2, 3])


**3D Grid of cubes**

>>> x = [0, 1, 0, 1, 0, 1, 0, 1]
>>> y = [0, 0, 1, 1, 0, 0, 1, 1]
>>> z = [0, 0, 0, 0, 1, 1, 1, 1]

>>> g = Structured(z, y, x, (2, 2, 2))

>>> g.get_x()
array([ 0.,  1.,  0.,  1.,  0.,  1.,  0.,  1.])
>>> g.get_y()
array([ 0.,  0.,  1.,  1.,  0.,  0.,  1.,  1.])
>>> g.get_z()
array([ 0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.])

>>> g.get_connectivity()
array([0, 1, 3, 2, 4, 5, 7, 6])
>>> g.get_offset()
array([8])

>>> g = Structured(x, y, z, (2, 2, 2), ordering='ccw')
>>> g.get_connectivity()
array([1, 0, 2, 3, 5, 4, 6, 7])
"""
import warnings

import numpy as np

from pymt.grids.connectivity import get_connectivity
from pymt.grids.utils import get_default_coordinate_names, get_default_coordinate_units

from .unstructured import Unstructured, UnstructuredPoints


class StructuredPoints(UnstructuredPoints):
    def __init__(self, *args, **kwds):
        """StructuredPoints(x0 [, x1 [, x2]], shape)"""
        (coordinates, shape) = (args[:-1], args[-1])

        if len(coordinates) < 1 or len(coordinates) > 3:
            raise ValueError("number of dimensions must be between 1 and 3")

        kwds.setdefault("set_connectivity", True)
        indexing = kwds.pop("indexing", "xy")

        if indexing != "ij":
            warnings.warn("only ij indexing is supported", RuntimeWarning)
            indexing = "ij"

        coordinate_names = kwds.pop(
            "coordinate_names", get_default_coordinate_names(len(coordinates))
        )
        coordinate_units = kwds.pop(
            "units", get_default_coordinate_units(len(coordinates))
        )

        self._shape = np.array(shape, dtype=np.int)

        kwds["units"] = coordinate_units
        kwds["coordinate_names"] = coordinate_names

        super(StructuredPoints, self).__init__(*coordinates, **kwds)

    def get_shape(self, remove_singleton=False):
        """The shape of the structured grid with the given indexing.

        Use remove_singleton=False to include singleton dimensions.
        """
        shape = self._shape.copy()
        if remove_singleton:
            shape = shape[shape > 1]

        return shape

    def get_coordinate_units(self, i):
        return super(StructuredPoints, self).get_coordinate_units(i)

    def get_coordinate_name(self, i):
        return super(StructuredPoints, self).get_coordinate_name(i)

    def get_coordinate(self, i):
        return super(StructuredPoints, self).get_coordinate(i)

    def get_point_coordinates(self, *args, **kwds):
        return super(StructuredPoints, self).get_point_coordinates(*args, **kwds)


class Structured(StructuredPoints, Unstructured):
    """Create a structured rectilinear grid.

    Parameters
    ----------
    x: ndarray
        1-D array of x-coordinates of nodes.
    y: ndarray
        1-D array of y-coordinates of nodes.
    shape: tuple of int
        Shape of the grid.
    indexing: {'xy', 'ij'}, optional
        Cartesian('xy', default) or matrix('ij') indexing of output.

    Returns
    -------
    Structured
        An instance of a Structured grid.
    """

    def __init__(self, *args, **kwds):
        kwds.setdefault("indexing", "xy")
        kwds.setdefault("set_connectivity", True)
        ordering = kwds.pop("ordering", "cw")
        if ordering not in ["cw", "ccw"]:
            raise TypeError("ordering not understood (valid choices are 'cw' or 'ccw')")

        shape = args[-1]

        if kwds["set_connectivity"]:
            (c, o) = get_connectivity(shape, ordering=ordering, with_offsets=True)
            self._set_connectivity(c, o)
            kwds["set_connectivity"] = False

        super(Structured, self).__init__(*args, **kwds)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#! /bin/env python

"""
>>> get_connectivity((6, ))
array([0, 1, 1, 2, 2, 3, 3, 4, 4, 5])

>>> get_connectivity((6, ), ordering='ccw')
array([1, 0, 2, 1, 3, 2, 4, 3, 5, 4])

>>> get_connectivity((3, 4))
array([ 0,  1,  5,  4,  1,  2,  6,  5,  2,  3,  7,  6,  4,  5,  9,  8,  5,
        6, 10,  9,  6,  7, 11, 10])

>>> get_connectivity((3, 4), ordering='ccw')
array([ 1,  0,  4,  5,  2,  1,  5,  6,  3,  2,  6,  7,  5,  4,  8,  9,  6,
        5,  9, 10,  7,  6, 10, 11])

>>> shape = np.array([3, 4, 4])
>>> (c, o) = get_connectivity(shape, with_offsets=True)

>>> len(c) == (shape - 1).prod() * 8
True
>>> len(o) == (shape - 1).prod()
True
>>> np.all(np.diff(o) == 8)
True

>>> c[:o[0]]
array([ 0,  1,  5,  4, 16, 17, 21, 20])
>>> c[o[-2]:o[-1]]
array([26, 27, 31, 30, 42, 43, 47, 46])

>>> ids = get_connectivity((3, 4, 4), ordering='ccw')
>>> ids[:8]
array([ 1,  0,  4,  5, 17, 16, 20, 21])
>>> ids[-8:]
array([27, 26, 30, 31, 43, 42, 46, 47])

"""

import numpy as np


def _get_interior_ids(shape, dtype=int):
    """
    Get indices to the interior nodes of a structured grid. These are
    essentially the upper-left corners of cells formed by the points.
    For a 1D grid this is the left node of each line segment. For a
    3D grid this is the upper-left corner of the closest face of each
    cube.

    >>> _get_interior_ids((6, ))
    array([0, 1, 2, 3, 4])
    >>> _get_interior_ids((3, 4))
    array([0, 1, 2, 4, 5, 6])
    >>> _get_interior_ids((2, 3, 4))
    array([0, 1, 2, 4, 5, 6])
    >>> _get_interior_ids((3, 3, 4))
    array([ 0,  1,  2,  4,  5,  6, 12, 13, 14, 16, 17, 18])
    """
    i = np.arange(np.prod(shape), dtype=dtype)
    i.shape = shape
    obj = []
    for d in shape:
        obj.append(slice(0, d - 1))
    return i[tuple(obj)].flatten()


def _get_offsets(shape, ordering="cw", dtype=int):
    """
    >>> _get_offsets((16, ))
    array([0, 1])
    >>> _get_offsets((3, 4))
    array([0, 1, 5, 4])
    >>> _get_offsets((3, 4, 4))
    array([ 0,  1,  5,  4, 16, 17, 21, 20])

    >>> _get_offsets((16, ), ordering='ccw')
    array([1, 0])
    >>> _get_offsets((3, 4), ordering='ccw')
    array([1, 0, 4, 5])
    >>> _get_offsets((3, 4, 4), ordering='ccw')
    array([ 1,  0,  4,  5, 17, 16, 20, 21])
    """

    if ordering not in ["cw", "ccw", "none"]:
        raise TypeError(
            "Ordering value not understood (known values are 'cw', 'ccw' and 'none')"
        )
    if ordering == "none":
        ordered = False
    else:
        ordered = True

    if ordering == "ccw":
        offsets = np.array([1, 0], dtype=dtype)
    else:
        offsets = np.array([0, 1], dtype=dtype)

    strides = np.cumprod(list(shape[-1:0:-1]))
    for (dim, stride) in enumerate(strides):
        new_offsets = offsets + stride
        if ordered and dim % 2 == 0:
            offsets = np.append(offsets, new_offsets[::-1])
        else:
            offsets = np.append(offsets, new_offsets)

    return offsets


def get_connectivity(shape, **kwds):
    """
    Get the connectivity (and, optionally, offset) array for an ND structured
    grid. Elements will consist of two points for 1D grids, four points for
    2D grids, eight points for 3D grids, etc. For a uniform rectilinear grid
    this would be lines, squares, and cubes.

    The nodes of an element can be ordered either clockwise or counter-clockwise.

    Parameters
    ----------
    shape: tuple of int
        The shape of the grid.
    ordering: {'cw', 'ccw', 'none'}, optional
        Node ordering.  One of 'cw' (clockwise), 'ccw' (counter-clockwise),
        or 'none' (ordering not guaranteed).
    dtype: str or numpy.dtype
        The desired data type of the returned arrays.
    with_offsets: bool
        Return offset array along with connectivity

    Returns
    -------
    ndarray of int
        Array of connectivities. If with_offsets keyword is True, return a
        tuple of (connectivity, offset).


    Examples
    --------

    A 1D grid with three points has 2 elements.

    >>> get_connectivity ((3, ))
    array([0, 1, 1, 2])
    >>> get_connectivity ((3, ), ordering='ccw')
    array([1, 0, 2, 1])


    A 2D grid with 3 rows and 4 columns of nodes::

        ( 0 ) --- ( 1 ) --- ( 2 ) --- ( 3 )
          |         |         |         |
          |    0    |    1    |    2    |
          |         |         |         |
        ( 4 ) --- ( 5 ) --- ( 6 ) --- ( 7 )
          |         |         |         |
          |    3    |    4    |    5    |
          |         |         |         |
        ( 8 ) --- ( 9 ) --- ( 10) --- ( 11)

    >>> get_connectivity((3, 4))
    array([ 0,  1,  5,  4,  1,  2,  6,  5,  2,  3,  7,  6,  4,  5,  9,  8,  5,
            6, 10,  9,  6,  7, 11, 10])

    >>> get_connectivity((3, 4), ordering='ccw')
    array([ 1,  0,  4,  5,  2,  1,  5,  6,  3,  2,  6,  7,  5,  4,  8,  9,  6,
            5,  9, 10,  7,  6, 10, 11])

    If ordering doesn't matter, set ordering to 'none' as this could be slightly faster.

    >>> (ids, offsets) = get_connectivity((3, 4), ordering='none', with_offsets=True)
    >>> offsets
    array([ 4,  8, 12, 16, 20, 24])
    >>> ids
    array([ 0,  1,  4,  5,  1,  2,  5,  6,  2,  3,  6,  7,  4,  5,  8,  9,  5,
            6,  9, 10,  6,  7, 10, 11])

    Nodes connected to the first cell,

    >>> ids[:offsets[0]]
    array([0, 1, 4, 5])

    Nodes connected to the second cell,

    >>> ids[offsets[0]:offsets[1]]
    array([1, 2, 5, 6])

    Instead of using an offset array to indicate the end of each cell,
    you can return a list of connectivity arrays for each cell.

    >>> shape = np.array((3, 4))
    >>> ids = get_connectivity(shape, ordering='cw', as_cell_list=True)
    >>> len(ids) == (shape - 1).prod()
    True
    >>> ids # doctest: +NORMALIZE_WHITESPACE
    [array([0, 1, 5, 4]),
     array([1, 2, 6, 5]),
     array([2, 3, 7, 6]),
     array([4, 5, 9, 8]),
     array([ 5,  6, 10,  9]),
     array([ 6,  7, 11, 10])]
    """
    kwds.setdefault("dtype", int)
    kwds.setdefault("ordering", "cw")
    with_offsets = kwds.pop("with_offsets", False)
    as_cell_list = kwds.pop("as_cell_list", False)

    if with_offsets and as_cell_list:
        raise ValueError("with_offsets and as_cell_list keywords cannot both be True")

    c0 = _get_interior_ids(shape, dtype=kwds["dtype"])
    offsets = _get_offsets(shape, **kwds)
    points_per_cell = len(offsets)

    if as_cell_list:
        offset_list = []
        for offset in offsets:
            offset_list.append(c0 + offset)
        c = []
        for cell in zip(*offset_list):
            c.append(np.array(cell, dtype=kwds["dtype"]))
    else:
        c = np.empty((c0.size * points_per_cell,), dtype=kwds["dtype"])
        for (i, offset) in enumerate(offsets):
            c[i::points_per_cell] = c0 + offset

    if with_offsets:
        o = np.arange(points_per_cell, len(c) + 1, points_per_cell, dtype=kwds["dtype"])
        return (c, o)
    else:
        return c


def get_connectivity_2d(shape, ordering="cw", dtype=int):
    """This is a little slower than the above and less general.

    Examples
    --------
    >>> get_connectivity_2d((3, 4))
    array([ 0,  1,  5,  4,  1,  2,  6,  5,  2,  3,  7,  6,  4,  5,  9,  8,  5,
            6, 10,  9,  6,  7, 11, 10])
    """
    if len(shape) != 2:
        raise ValueError("number of dimensions must be 2")

    point_count = shape[0] * shape[1]

    point_ids = np.arange(point_count, dtype=dtype)
    point_ids.shape = shape

    c_ul = point_ids[:-1, :-1].flatten()
    c_ur = point_ids[:-1, 1:].flatten()
    c_lr = point_ids[1:, 1:].flatten()
    c_ll = point_ids[1:, :-1].flatten()

    c = np.empty((4 * c_ul.size,), dtype=c_ul.dtype)
    if ordering == "cw":
        c[::4] = c_ul
        c[1::4] = c_ur
        c[2::4] = c_lr
        c[3::4] = c_ll
    else:
        c[::4] = c_ur
        c[1::4] = c_ul
        c[2::4] = c_ll
        c[3::4] = c_lr

    return c


def get_connectivity_1d(shape, ordering="cw", dtype=int):
    if len(shape) != 1:
        raise ValueError("number of dimensions must be 1")

    point_count = shape[0]

    point_ids = np.arange(point_count, dtype=dtype)

    c_left = point_ids[:-1]
    c_right = point_ids[1:]

    c = np.empty((4 * c_left.size,), dtype=c_left.dtype)
    if ordering == "cw":
        c[::2] = c_left
        c[1::2] = c_right
    else:
        c[::2] = c_right
        c[1::2] = c_left

    return c


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



import collections
import re

import six

_TIME_SERIES_NAME_RE_PATTERN = r"(?P<name>[a-zA-Z_]+)@t=(?P<time_stamp>\d+)$"
_TIME_SERIES_NAME_RE = re.compile(_TIME_SERIES_NAME_RE_PATTERN)


class Error(Exception):
    pass


class TimeSeriesNameError(Error):
    def __init__(self, name):
        self._name = name

    def __str__(self):
        return "invalid time series name: %s" % self._name


class TimeStampError(Error):
    def __init__(self, time_stamp):
        self._time_stamp = time_stamp

    def __str__(self):
        return "invalid time stamp: %s" % self._time_stamp


def time_stamp_as_string(time_stamp):
    if isinstance(time_stamp, six.string_types):
        try:
            return str(int(time_stamp))
        except ValueError:
            raise TimeStampError(time_stamp)
    elif isinstance(time_stamp, six.integer_types):
        return str(time_stamp)
    else:
        raise TimeStampError(time_stamp)


def split(name):
    match = _TIME_SERIES_NAME_RE.match(name)
    if match is not None:
        return match.group("name"), int(match.group("time_stamp"))
    else:
        raise TimeSeriesNameError(name)


def unsplit(name, time_stamp):
    return name + "@t=" + time_stamp_as_string(time_stamp)


def extract_time_stamps_from_names(names, ordering="ascending", prefix=""):
    if ordering not in ["ascending", "descending", None]:
        raise TypeError("ordering not understood: %s" % ordering)

    time_stamps = collections.defaultdict(list)

    for name in names:
        if name.startswith(prefix):
            try:
                (base, time_stamp) = split(name)
            except TimeSeriesNameError:
                pass
            else:
                time_stamps[base].append(time_stamp)

    if ordering is not None:
        for name in time_stamps:
            time_stamps[name].sort(reverse=(ordering == "descending"))

    return time_stamps


def sort_time_series_names(names, ordering="ascending", prefix=""):
    ordered_names = dict()

    time_stamps = extract_time_stamps_from_names(
        names, ordering=ordering, prefix=prefix
    )
    for (name, stamps) in time_stamps.items():
        ordered_names[name] = [unsplit(name, stamp) for stamp in stamps]

    return ordered_names


def get_time_series_names(names):
    unique_names = set()

    for name in names:
        try:
            unique_names.add(split(name)[0])
        except TimeSeriesNameError:
            pass

    return unique_names



#! /bin/env python
"""PyMT interface to ESMPy grids.

**Examples**

Create a grid that looks like this,

::

    (3) --- (4) --- (5)
     |       |       |
     |   0   |   1   |
     |       |       |
    (0) --- (1) --- (2)


>>> import ESMF
>>> mngr = ESMF.Manager()
>>> mngr.local_pet
0
>>> mngr.pet_count
1

Represent this grid as an unstructured grid.

>>> g = EsmpUnstructured([0, 0, 0, 1, 1, 1], [0, 1, 2, 0, 1, 2],
...                      [0, 1, 4, 3, 1, 2, 5, 4], [4, 8], indexing='ij')
>>> g.get_point_count()
6
>>> g.get_cell_count()
2

As a structured grid,

>>> g = EsmpStructured([0, 0, 0, 1, 1, 1], [0, 1, 2, 0, 1, 2], (2, 3),
...                    indexing='ij')
>>> g.get_point_count()
6
>>> g.get_cell_count()
2

As a uniform rectilinear (raster) grid,

>>> g = EsmpUniformRectilinear([2, 3], [1., 1.], [0., 0.])
>>> g.get_point_count()
6
>>> g.get_cell_count()
2

The as_mesh method provides a view of the grid as an ESMP_Mesh.

>>> mesh = g.as_mesh()
>>> mesh.node_count
6
>>> mesh.element_count
2

ESMF elements are the same as the grids cells. Likewise with nodes and points.

>>> g = EsmpRectilinear([0, 1], [0, 1, 2], indexing='ij')
>>> g.as_mesh().element_count == g.get_cell_count()
True
>>> g.as_mesh().node_count == g.get_point_count()
True


**Uniform Rectilinear Field**

Create a field on a grid that looks like this,

::

    (3) --- (4) --- (5)
     |       |       |
     |   0   |   1   |
     |       |       |
    (0) --- (1) --- (2)

Create the field,

>>> g = EsmpRasterField((2, 3), (1, 2), (0, 0), indexing='ij')
>>> g.get_cell_count()
2
>>> g.get_point_count()
6

Add some data at the points of our grid.

>>> data = np.arange(6)
>>> g.add_field('var0', data, centering='point')
>>> f = g.get_field('var0')
>>> f.data
array([ 0.,  1.,  2.,  3.,  4.,  5.])

The data can be given either as a 1D array or with the same shape
as the point grid. In either case, though, it will be flattened.

>>> data = np.arange(6)
>>> data.shape = (2, 3)
>>> g.add_field('var0', data, centering='point')
>>> f = g.get_field('var0')
>>> f.data
array([ 0.,  1.,  2.,  3.,  4.,  5.])

If the size or shape doesn't match, it's an error.

>>> data = np.arange(2)
>>> g.add_field('bad var', data, centering='point') # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
    ...
DimensionError: 2 != 6

>>> data = np.ones((3, 2))
>>> g.add_field ('bad var', data, centering='point') # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
    ...
DimensionError: (3, 2) != (2, 3)


**Map between two fields**

>>> from pymt.grids.raster import UniformRectilinear
>>> from pymt.grids.rectilinear import Rectilinear

>>> src = EsmpRasterField((3,3), (1,1), (0, 0), indexing='ij')
>>> data = np.arange(src.get_cell_count(), dtype=np.float64)
>>> src.add_field('srcfield', data, centering='zonal')
>>> src.get_point_count()
9
>>> src.get_cell_count()
4
>>> src.get_x()
array([ 0.,  1.,  2.,  0.,  1.,  2.,  0.,  1.,  2.])
>>> src.get_y()
array([ 0.,  0.,  0.,  1.,  1.,  1.,  2.,  2.,  2.])
>>> src.get_connectivity() + 1
array([1, 2, 5, 4, 2, 3, 6, 5, 4, 5, 8, 7, 5, 6, 9, 8])

>>> dst = EsmpRectilinearField([0., .5, 1.5, 2.], [0., .5, 1.5, 2.])
>>> data = np.empty(dst.get_cell_count(), dtype=np.float64)
>>> dst.add_field('dstfield', data, centering='zonal')
>>> dst.get_point_count()
16
>>> dst.get_cell_count()
9
>>> dst.get_x()
array([ 0. ,  0.5,  1.5,  2. ,  0. ,  0.5,  1.5,  2. ,  0. ,  0.5,  1.5,
        2. ,  0. ,  0.5,  1.5,  2. ])
>>> dst.get_y()
array([ 0. ,  0. ,  0. ,  0. ,  0.5,  0.5,  0.5,  0.5,  1.5,  1.5,  1.5,
        1.5,  2. ,  2. ,  2. ,  2. ])
>>> dst.get_connectivity() + 1
array([ 1,  2,  6,  5,  2,  3,  7,  6,  3,  4,  8,  7,  5,  6, 10,  9,  6,
        7, 11, 10,  7,  8, 12, 11,  9, 10, 14, 13, 10, 11, 15, 14, 11, 12,
       16, 15])

>>> src_field = src.as_esmp('srcfield')
>>> dst_field = dst.as_esmp('dstfield')
>>> src.as_mesh().element_count
4
>>> src.as_mesh().node_count
9
>>> dst.as_mesh().element_count
9
>>> dst.as_mesh().node_count
16

>>> f = run_regridding(src_field, dst_field)
>>> f.data
array([ 0. ,  0.5,  1. ,  1. ,  1.5,  2. ,  2. ,  2.5,  3. ])

**A bigger grid**

>>> (M, N) = (300, 300)
>>> src = EsmpRasterField((M, N), (1, 1), (0, 0))

**Map values on cells**

>>> (X, Y) = np.meshgrid(np.arange (0.5, 299.5, 1.),
...                      np.arange (0.5, 299.5, 1.))
>>> data = np.sin(np.sqrt(X ** 2 + Y ** 2) * np.pi / M)
>>> src.add_field('srcfield', data, centering='zonal')

>>> dst = EsmpRasterField((M * 2 - 1, N * 2 - 1), (1. / 2, 1. / 2), (0, 0))
>>> data = np.empty(dst.get_cell_count(), dtype=np.float64)
>>> dst.add_field('dstfield', data, centering='zonal')

>>> src_field = src.as_esmp('srcfield')
>>> dst_field = dst.as_esmp('dstfield')

>>> f = run_regridding(src_field, dst_field)

>>> (X, Y) = np.meshgrid(np.arange (0.5, 299.5, .5),
...                      np.arange (0.5, 299.5, .5))
>>> exact = np.sin(np.sqrt(X ** 2 + Y ** 2) * np.pi / M)
>>> np.sum(np.abs(exact.flat - f.data))/(M * N * 4.) < 1e-2
True

**Map values on points**

>>> (X, Y) = np.meshgrid(np.arange(0.5, 300.5, 1.),
...                      np.arange(0.5, 300.5, 1.))
>>> data = np.sin(np.sqrt(X ** 2 + Y ** 2) * np.pi / M)
>>> src.add_field('srcfield_at_points', data, centering='point')

>>> data = np.empty(dst.get_point_count(), dtype=np.float64)
>>> dst.add_field('dstfield_at_points', data, centering='point')

>>> src_field = src.as_esmp('srcfield_at_points')
>>> dst_field = dst.as_esmp('dstfield_at_points')

>>> f = run_regridding(src_field, dst_field,
...                    method=ESMF.RegridMethod.BILINEAR)

>>> (X, Y) = np.meshgrid(np.arange(0.5, 300., .5), np.arange(0.5, 300., .5))
>>> exact = np.sin(np.sqrt (X ** 2 + Y ** 2) * np.pi / M)
>>> np.sum(np.abs(exact.flat - f.data))/(M * N * 4.) < 1e-5
True
"""

import numpy as np

from pymt.grids import Rectilinear, Structured, UniformRectilinear, Unstructured
from pymt.grids.igrid import (
    CENTERING_CHOICES,
    CenteringValueError,
    DimensionError,
    IField,
    IGrid,
)

try:
    import ESMF as esmf
except ImportError:
    esmf = None
    __doc__ = (  # pylint: disable=redefined-builtin
        "This module is not available (no ESMF installation was found)"
    )


class EsmpGrid(IGrid):
    def __init__(self):
        self._mesh = esmf.Mesh(parametric_dim=2, spatial_dim=2)

        self._mesh_add_nodes()
        self._mesh_add_elements()

        super(EsmpGrid, self).__init__()

    def get_point_count(self):
        raise NotImplementedError("get_point_count")

    def get_cell_count(self):
        raise NotImplementedError("get_cell_count")

    def as_mesh(self):
        return self._mesh

    def _mesh_add_nodes(self):
        node_ids = np.arange(1, self.get_point_count() + 1, dtype=int)
        (x, y) = (self.get_x(), self.get_y())

        node_coords = np.empty(x.size + y.size, dtype=np.float64)
        (node_coords[0::2], node_coords[1::2]) = (x, y)

        node_owner = np.zeros(self.get_point_count(), dtype=int)

        self._mesh.add_nodes(self.get_point_count(), node_ids, node_coords, node_owner)

    def _mesh_add_elements(self):
        cell_ids = np.arange(1, self.get_cell_count() + 1, dtype=int)
        cell_types = np.empty(self.get_cell_count(), dtype=int)
        cell_types.fill(esmf.MeshElemType.QUAD)

        cell_conn = np.array(self.get_connectivity(), dtype=int)  # + 1

        self._mesh.add_elements(self.get_cell_count(), cell_ids, cell_types, cell_conn)


class EsmpUnstructured(Unstructured, EsmpGrid):
    name = "ESMPUnstructured"


class EsmpStructured(Structured, EsmpGrid):
    name = "ESMPStructured"


class EsmpRectilinear(Rectilinear, EsmpGrid):
    name = "ESMPRectilinear"


class EsmpUniformRectilinear(UniformRectilinear, EsmpStructured):
    name = "ESMPUniformRectilinear"


class EsmpField(IField):
    def __init__(self, *args, **kwargs):
        super(EsmpField, self).__init__(*args, **kwargs)
        self._fields = {}

    def get_point_count(self):
        raise NotImplementedError("get_point_count")

    def get_cell_count(self):
        raise NotImplementedError("get_cell_count")

    def add_field(self, field_name, val, centering="zonal"):
        if centering not in CENTERING_CHOICES:
            raise CenteringValueError(centering)

        if centering == "zonal" and val.size != self.get_cell_count():
            raise DimensionError(val.size, self.get_cell_count())
        elif centering != "zonal" and val.size != self.get_point_count():
            raise DimensionError(val.size, self.get_point_count())

        if centering == "zonal":
            meshloc = esmf.MeshLoc.ELEMENT
        else:
            meshloc = esmf.MeshLoc.NODE

        field = esmf.Field(self._mesh, field_name, meshloc=meshloc)
        np.copyto(field.data, val.view().reshape(field.data.shape))
        self._fields[field_name] = field

    def get_field(self, field_name):
        return self._fields[field_name]

    def as_esmp(self, field_name):
        return self._fields[field_name]


class EsmpStructuredField(EsmpStructured, EsmpField):
    def add_field(self, field_name, val, centering="zonal"):
        if centering == "zonal":
            if val.ndim > 1 and np.any(val.shape != self.get_shape() - 1):
                raise DimensionError(val.shape, self.get_shape() - 1)
        elif centering != "zonal":
            if val.ndim > 1 and np.any(val.shape != self.get_shape()):
                raise DimensionError(val.shape, self.get_shape())
        try:
            super(EsmpStructuredField, self).add_field(
                field_name, val, centering=centering
            )
        except (DimensionError, CenteringValueError):
            raise


class EsmpUnstructuredField(EsmpUnstructured, EsmpField):
    pass


class EsmpRectilinearField(EsmpRectilinear, EsmpStructuredField):
    pass


class EsmpRasterField(EsmpUniformRectilinear, EsmpRectilinearField):
    pass


def run_regridding(srcfield, dstfield, **kwds):
    """run_regridding(source_field, destination_field, method=ESMP_REGRIDMETHOD_CONSERVE, unmapped=ESMP_UNMAPPEDACTION_ERROR)

    **PRECONDITIONS:**
        Two ESMP_Fields have been created and a regridding operation is desired from 'srcfield' to 'dstfield'.

    **POSTCONDITIONS:**
        An ESMP regridding operation has set the data on 'dstfield'.
    """
    method = kwds.get("method", esmf.RegridMethod.CONSERVE)
    unmapped = kwds.get("unmapped", esmf.UnmappedAction.ERROR)

    # call the regridding functions
    regridder = esmf.Regrid(
        srcfield, dstfield, regrid_method=method, unmapped_action=unmapped
    )
    dstfield = regridder(srcfield, dstfield)

    return dstfield



#! /usr/bin/env python
import jinja2
import six
from landlab.core.messages import format_message
from model_metadata import MetadataNotFoundError, ModelMetadata


# {{ desc|trim|wordwrap(70) if desc }}
_DOCSTRING = u"""
Basic Model Interface for {{ name }}.

{{ desc|trim if desc }}

{% if author -%}
Author:
{%- for name in author %}
{{"- " + name|trim }}{% endfor %}
{%- endif %}
Version: {{ version }}
License: {{ license }}
DOI: {{ doi }}
URL: {{ url }}

{% if cite_as -%}
Cite as:
{% for citation in cite_as %}
{{ citation|trim|indent(width=4, first=True) }}
{% endfor %}
{%- endif %}
{% if parameters %}
Parameters
----------
{% for param in parameters|sort(attribute='name') -%}
{{param.name}} : {{param.value.type}}, optional
    {{ "%s [default=%s %s]"|format(param.description, param.value.default, param.value.units)|trim|wordwrap(70)|indent(4) }}
{% endfor %}
{% endif -%}

Examples
--------
>>> from pymt.models import {{name}}
>>> model = {{name}}()
>>> (fname, initdir) = model.setup()
>>> model.initialize(fname, dir=initdir)  # doctest: +SKIP
>>> for _ in range(10):  # doctest: +SKIP
...     model.update()
>>> model.finalize()  # doctest: +SKIP
""".strip()


def bmi_docstring(
    plugin,
    author=None,
    version=None,
    license=None,  # pylint: disable=redefined-builtin
    doi=None,
    url=None,
    parameters=None,
    summary=None,
    cite_as=None,
    email=None,
):
    """Build the docstring for a BMI model.

    Parameters
    ----------
    name : str
        Name of a BMI component.
    author : str, optional
        Name of author or authors.
    version : str, optional
        Version string for the component.
    license : str, optional
        Name of the license of the component.
    doi : str, optional
        A DOI for the component.
    url : str, optional
        URL of the component's location on the internet.
    parameters : iterable, optional
        List of input parameters for the component. Each parameter object
        must have attributes for name, type, value, units, and desc.
    cite_as : iterable of str, optional
        List of citations for this component.
    email : str, optional
        Contact email address.

    Returns
    -------
    str
        The docstring.

    Examples
    --------
    >>> from pymt.framework.bmi_docstring import bmi_docstring
    >>> print(bmi_docstring('Model', author='Walt Disney')) #doctest: +ELLIPSIS
    Basic Model Interface for Model.
    ...
    """
    author = author or []
    cite_as = cite_as or []

    try:
        meta = ModelMetadata.from_obj(plugin)
        # meta = PluginMetadata(plugin)
    except MetadataNotFoundError:
        if isinstance(plugin, six.string_types):
            info = dict(
                authors=author,
                version=version,
                license=license,
                doi=doi,
                url=url,
                summary=summary,
                cite_as=cite_as,
            )
            defaults = {}
            name = plugin
        else:
            raise
    else:
        info = meta.info
        defaults = meta.parameters
        name = meta.api["class"]

    for param_name, param in defaults.items():
        param["name"] = param_name
    defaults = defaults.values()

    author = author or info["authors"]
    email = email or "-"
    version = version or info["version"]
    license = license or info["license"]
    doi = doi or info["doi"]
    url = url or info["url"]
    summary = summary or info["summary"]
    cite_as = cite_as or info["cite_as"]
    parameters = parameters or defaults

    if isinstance(author, six.string_types):
        author = [author]
    if isinstance(cite_as, six.string_types):
        cite_as = [cite_as]

    summary = format_message(summary)

    env = jinja2.Environment(  # nosec
        loader=jinja2.DictLoader({"docstring": _DOCSTRING})
    )
    return env.get_template("docstring").render(
        desc=summary,
        name=name,
        parameters=parameters,
        author=author,
        version=version,
        license=license,
        doi=doi,
        url=url,
        cite_as=cite_as,
    )



"""
Examples
--------
Create an instance of the *AirPort* component and run it. The `go`
method will initialize the component, run it for
its duration, and finalize it.

>>> from pymt.framework.services import register_component_classes
>>> register_component_classes(['pymt.testing.services.AirPort'])
>>> comp = Component('AirPort')
>>> comp.start_time
0.0
>>> comp.current_time
0.0
>>> comp.end_time
100.0
>>> comp.go() # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
{name: air_port, status: running, time: 1.0}
...
{name: air_port, status: running, time: 100.0}
>>> comp.current_time
100.0

If you try to create a new component with the same name, it will raise
an exception. To create a new instance of the component, you have to
give it a new name.

>>> comp = Component('AirPort') # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
ValueError: AirPort
>>> comp = Component('AirPort', name='air_port')
>>> comp.current_time
0.0

You can gain finer control over component execution with the
`initialize`, `run`, and `finalize` methods.

>>> comp.initialize()
>>> comp.run(10.) # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
{name: air_port, status: running, time: 1.0}
...
{name: air_port, status: running, time: 10.0}
>>> comp.current_time
10.0
>>> comp.run(101.) # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
ValueError: AirPort
>>> comp.current_time
10.0
>>> comp.run(100.) # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
{name: air_port, status: running, time: 11.0}
...
{name: air_port, status: running, time: 100.0}
>>> comp.current_time
100.0
>>> comp.finalize()
"""
import warnings

import six
import yaml

from ..events.chain import ChainEvent
from ..events.manager import EventManager
from ..events.port import PortEvent, PortMapEvent
from ..events.printer import PrintEvent
from ..framework import services
from .grid import GridMixIn


def clip_stop_time(stop, stop_min, stop_max):
    """Clip time between two values.

    Clip a stopping time between two values. This is the same as the numpy
    clip function except that if *stop* is `None` then *stop_max* is
    returned.

    Parameters
    ----------
    stop : float
        Stop time.
    stop_min : float
        Minimum stop time
    stop_max : float
        Maximum stop time

    Returns
    -------
    float
        The, possibly clipped, stop time.

    Examples
    --------
    >>> clip_stop_time(1., 0, 2.)
    1.0
    >>> clip_stop_time(1., 2, 3)
    2.0
    >>> clip_stop_time(4., 2, 3)
    3.0
    >>> clip_stop_time(None, 0, 1)
    1.0
    """
    from numpy import clip

    if stop is None:
        return float(stop_max)
    else:
        return clip(stop, stop_min, stop_max)


class Component(GridMixIn):
    """Wrap a BMI object as a component.

    Use the Component class to wrap an object that exposes a BMI so that it
    can operate within the model-coupling framework.

    Parameters
    ----------
    port : str or port-like
        Name of the port to wrap or a Port instance.
    uses : list, optional
        List of uses port names.
    provides : list, optional
        List of provides port names.
    argv : tuple, optional
        List of command line argument used to initialize the component.
    time_step : float, optional
        Time interval over which component will run uses ports.
    run_dir : str, optional
        Directory where the component will run.
    """

    def __init__(
        self,
        port,
        uses=None,
        provides=None,
        events=(),
        argv=(),
        time_step=1.0,
        run_dir=".",
        name=None,
    ):
        if uses is None:
            uses = set()
        if provides is None:
            provides = set()

        if isinstance(port, six.string_types):
            if name is None:
                name = port
            self._port = services.instantiate_component(port, name)
        else:
            self._port = port
        self._uses = uses
        self._provides = provides
        self._time_step = time_step

        self._events = EventManager(
            [(PortEvent(port=self._port, init_args=argv, run_dir=run_dir), time_step)]
            + list(events)
        )
        # self._events = EventManager([(self._port, 1.)] + events)
        # self._events = EventManager(
        #    [(self._port, self._port.time_step)] + events)

        GridMixIn.__init__(self)

    @property
    def time_step(self):
        """Component time step.

        Time step over which a component will update any connected uses ports.
        """
        return self._time_step

    @property
    def current_time(self):
        """Current time for component updating."""
        try:
            return self._port.current_time
        except AttributeError:
            return self._port.get_current_time()

    @property
    def start_time(self):
        """Start time for component updating."""
        try:
            return self._port.start_time
        except AttributeError:
            return self._port.get_start_time()

    @property
    def end_time(self):
        """End time for component updating."""
        try:
            return self._port.end_time
        except AttributeError:
            return self._port.get_end_time()

    @property
    def uses(self):
        """Names of connected *uses* ports."""
        return self._uses

    @property
    def provides(self):
        """Names of connected *provides* ports."""
        return self._provides

    def connect(self, uses, port, vars_to_map=()):
        """Connect a *uses* port to a *provides* port.

        Parameters
        ----------
        uses : str
            Name of the *uses* port.
        port : port-like
            Port-like object to connect to.
        var_to_map : iterable, optional
            Names of variables to map.
        """
        if uses not in self.uses:
            warnings.warn("component does not use %s" % uses)

        if len(vars_to_map) > 0:
            event = ChainEvent(
                [
                    port,
                    PortMapEvent(src_port=port, dst_port=self, vars_to_map=vars_to_map),
                ]
            )
            # PortMapEvent(src_port=port, dst_port=self._port,
        else:
            event = port
        self._events.add_recurring_event(event, port.time_step)
        # self._events.add_recurring_event(event, port._port.time_step)
        # self._events.add_recurring_event(port, port._port.time_step)

    def go(self, stop=None):
        """Run a component from start to end.

        Run a component starting from its start time and ending at its stop
        time. The component and any attached ports are first initialized,
        then updated until the end time and, finally, their finalize methods
        are called.

        Parameters
        ----------
        stop : float, optional
            Stop time, or None to run until `end_time`.
        """
        self.initialize()

        stop_time = clip_stop_time(stop, self.time_step, self.end_time)
        try:
            self.run(stop_time)
        except Exception:
            raise
        finally:
            self._events.finalize()

    def initialize(self):
        """Initialize a component and any connected events."""
        self._events.initialize()

    def run(self, stop_time):
        """Run a component and any connect events.

        Parameters
        ----------
        stop_time : float
            Time to run the component until.
        """
        if stop_time > self.end_time:
            raise ValueError("stop time is greater than end time.")
        self._events.run(stop_time)

    def finalize(self):
        """Finalize a component and any connected events."""
        self._events.finalize()

    def register(self, name):
        """Register component with the framework.

        Associate *name* with the component instance withing the framework.

        Parameters
        ----------
        name : str
            Name to associate component with.
        """
        services.register_component_instance(name, self)

    @classmethod
    def from_string(cls, source):
        """Deprecated.

        .. note:: Deprecated.
            Use :meth:`~Component.load` instead.

        Create a component from a string.

        Parameters
        ----------
        source : str
            Contents of a yaml file.

        Returns
        -------
        Component
            A newly-created component.
        """
        return cls._from_yaml(source)

    @classmethod
    def load(cls, source):
        """Create a Component from a string.

        This is an alternate constructor that create a component using values
        from a yaml-formatted string.

        Parameters
        ----------
        source : str
            Yaml-formatted string.

        Returns
        -------
        Component
            A newly-created component.
        """
        return cls.from_dict(yaml.safe_load(source))

    @classmethod
    def load_all(cls, source):
        """Create multiple components from a string.

        This is an alternate constructor that creates a series of components
        using values from a yaml-formatted string.

        Parameters
        ----------
        source : str
            Yaml-formatted string.

        Returns
        -------
        List of Components
            A list of newly-created component.
        """
        components = []
        for section in yaml.safe_load_all(source):
            components.append(cls.from_dict(section))
        return components

    @classmethod
    def from_dict(cls, d):
        """Create a Component instance from a dictionary.

        Use configuration paramters from a dict-like object to create a new
        Component. The dictionary must contain the following keys:
        * `name`
        * `class`
        * `initialize_args`
        * `time_step`
        * `run_dir`

        Parameters
        ----------
        d : dict-like
            Configuration parameters for the component.

        Returns
        -------
        Component
            A newly-created Component instance.
        """
        # port = services.get_port(d['name'])
        port = services.instantiate_component(d["class"], d["name"])
        # argv = d.get('argv', [])
        try:
            argv = d["initialize_args"]
        except KeyError:
            argv = d.get("argv", [])
        time_step = float(d.get("time_step", 1.0))
        run_dir = d.get("run_dir", ".")

        events = []
        for conf in d.get("print", []):
            conf["port"] = port
            conf["run_dir"] = run_dir
            events.append((PrintEvent(**conf), conf["interval"]))

        uses = d.get("uses", [])
        provides = d.get("provides", [])

        return cls(
            port,
            uses=uses,
            provides=provides,
            events=events,
            argv=argv,
            time_step=time_step,
            run_dir=run_dir,
        )

    @classmethod
    def from_path(cls, path):
        """Create a component from a file.

        Parameters
        ----------
        path : str
            Path to yaml file.

        Returns
        -------
        Component
            A newly-created component.
        """
        with open(path, "r") as yaml_file:
            return cls._from_yaml(yaml_file.read())

    @classmethod
    def _from_yaml(cls, contents):
        comp = yaml.safe_load(contents)
        return cls.from_dict(comp)



#! /usr/bin/env python
import warnings

import numpy as np

try:
    import ESMF as esmf
except ImportError:
    esmf = None


REGRID_METHODS = {}
UNMAPPED_ACTIONS = {}

if esmf is not None:
    REGRID_METHODS.update(
        {
            "bilinear": esmf.RegridMethod.BILINEAR,
            "nearest": esmf.RegridMethod.NEAREST_STOD,
            "conserve": esmf.RegridMethod.CONSERVE,
        }
    )
    UNMAPPED_ACTIONS.update(
        {"pass": esmf.UnmappedAction.IGNORE, "raise": esmf.UnmappedAction.ERROR}
    )


def ravel_jaggedarray(array):
    values_per_row = np.sum(array >= 0, axis=1)

    raveled_array = np.empty(values_per_row.sum(), dtype=array.dtype)
    offset = 0
    for row, n_values in zip(array, values_per_row):
        raveled_array[offset : offset + n_values] = row[row >= 0]
        offset += n_values

    return raveled_array, values_per_row


def bmi_as_esmf_mesh(bmi_grid):
    xy_at_node = np.vstack((bmi_grid.node_x.values, bmi_grid.node_y.values)).T.copy()

    if "face_node_connectivity" in bmi_grid:
        nodes_at_patch = bmi_grid.face_node_connectivity.values
        nodes_per_patch = np.diff(
            np.concatenate(([0], bmi_grid.face_node_offset.values))
        )
    else:
        nodes_at_patch = None
        nodes_per_patch = None

    return as_esmf_mesh(xy_at_node, nodes_at_patch, nodes_per_patch)

    # return as_esmf_mesh(xy_at_node, np.astype(nodes_at_patch, dtype=np.int32),
    #                     np.astype(nodes_per_patch, dtype=np.int32))


def as_esmf_mesh(xy_of_node, nodes_at_patch=None, nodes_per_patch=None):
    mesh = esmf.Mesh(parametric_dim=2, spatial_dim=2)

    n_nodes = len(xy_of_node)

    node_ids = np.arange(1, n_nodes + 1, dtype=np.int32)
    node_owner = np.zeros(n_nodes, dtype=np.int32)

    mesh.add_nodes(n_nodes, node_ids, xy_of_node, node_owner)

    if nodes_at_patch is not None:
        if nodes_at_patch.ndim == 2:
            n_faces = len(nodes_at_patch)
        else:
            n_faces = len(nodes_per_patch)

        face_ids = np.arange(1, n_faces + 1, dtype=np.int32)

        if nodes_at_patch.ndim == 2:
            face_conn, nodes_per_face = ravel_jaggedarray(nodes_at_patch)
        else:
            face_conn, nodes_per_face = nodes_at_patch, nodes_per_patch

        face_conn = face_conn.astype(dtype=np.int32, copy=False)
        nodes_per_face = nodes_per_face.astype(dtype=np.int32, copy=False)

        if np.all((nodes_per_face == 3) | (nodes_per_face == 4)):
            face_types = np.full(n_faces, -1, dtype=np.int32)
            face_types[np.where(nodes_per_face == 3)] = esmf.MeshElemType.TRI
            face_types[np.where(nodes_per_face == 4)] = esmf.MeshElemType.QUAD

            mesh.add_elements(n_faces, face_ids, face_types, face_conn)
        else:
            warnings.warn("mesh contains non-triangular or quadrilateral elements")

    return mesh


def as_esmf_field(mesh, field_name, data=None, at="node"):
    if at == "node":
        meshloc = esmf.MeshLoc.NODE
    elif at == "cell":
        meshloc = esmf.MeshLoc.ELEMENT
    else:
        raise ValueError("'at' location not understood (must be 'cell' or 'node')")

    field = esmf.Field(mesh, field_name, meshloc=meshloc)
    if data is not None:
        np.copyto(field.data, data.reshape(field.data.shape))

    return field


def graph_as_esmf(graph, field_name, data=None, at="node"):
    mesh = as_esmf_mesh(graph.xy_of_node, graph.nodes_at_patch)
    field = as_esmf_field(mesh, field_name, data=data, at=at)

    return field


def run_regridding(srcfield, dstfield, method="nearest", unmapped="pass"):
    """run_regridding(source_field, destination_field, method=ESMP_REGRIDMETHOD_CONSERVE, unmapped=ESMP_UNMAPPEDACTION_ERROR)

    **PRECONDITIONS:**
        Two ESMP_Fields have been created and a regridding operation is desired from 'srcfield' to 'dstfield'.
    **POSTCONDITIONS:**
        An ESMP regridding operation has set the data on 'dstfield'.
    """
    # method = kwds.get('method', ESMF.RegridMethod.NEAREST_STOD)
    # method = kwds.get('method', ESMF.RegridMethod.BILINEAR)
    # unmapped = kwds.get('unmapped', ESMF.UnmappedAction.IGNORE)
    # method = kwds.get('method', ESMF.RegridMethod.CONSERVE)
    # unmapped = kwds.get('unmapped', ESMF.UnmappedAction.ERROR)

    try:
        method = REGRID_METHODS[method]
    except KeyError:
        raise ValueError("regrid method not understood")
    try:
        unmapped = UNMAPPED_ACTIONS[unmapped]
    except KeyError:
        raise ValueError("unmapped action not understood")

    # call the regridding functions
    masked_values = np.array([-9999.0])
    regridder = esmf.Regrid(
        srcfield,
        dstfield,
        regrid_method=method,
        unmapped_action=unmapped,
        src_mask_values=masked_values,
        dst_mask_values=masked_values,
    )
    dstfield = regridder(srcfield, dstfield)

    return dstfield


class GridMapperMixIn(object):
    def _esmf_mesh_by_id(self, gid):
        try:
            self._esmf_mesh
        except AttributeError:
            self._esmf_mesh = dict()

        try:
            self._esmf_mesh[gid]
        except KeyError:
            self._esmf_mesh[gid] = bmi_as_esmf_mesh(self.grid[gid])

        return self._esmf_mesh[gid]

    def _esmf_field_by_id(self, gid, name=None, at="node"):
        name = name or "generic"

        try:
            self._esmf_field
        except AttributeError:
            self._esmf_field = dict()

        _id = "{gid}.{name}@{at}".format(gid=gid, name=name, at=at)

        try:
            self._esmf_field[_id]
        except KeyError:
            self._esmf_field[_id] = as_esmf_field(
                self._esmf_mesh_by_id(gid), name, at=at
            )

        return self._esmf_field[_id]

    def regrid(self, name, **kwds):
        """Regrid values from one grid to another.

        Parameters
        ----------
        name : str
            Name of the values to regrid.
        to : bmi_like, optional
            BMI object onto which to map values. If not provided, map
            values onto one of the object's own grids.
        to_name : str, optional
            Name of the value to map onto. If not provided, use *name*.

        Returns
        -------
        ndarray
            The regridded values.
        """
        dst = kwds.pop("to", self)
        dst_name = kwds.pop("to_name", name)

        data = self.get_value(name, **kwds)

        if esmf is not None:
            src_field = self._esmf_field_by_id(self.var[name].grid, at="node")
            dst_field = dst._esmf_field_by_id(dst.var[dst_name].grid, at="node")

            np.copyto(src_field.data, data.reshape(src_field.data.shape))

            run_regridding(src_field, dst_field)

            return dst_field.data
        else:
            return data

    def map_to(self, name, **kwds):
        """Map values to another grid.

        Parameters
        ----------
        name : str
            Name of values to push.
        """
        destination = kwds.pop("destination", self)
        at = kwds.pop("at", name)
        data = self.regrid(name, to=destination, to_name=at, **kwds)
        destination.set_value(at, data)

    def set_value(self, name, *args, **kwds):
        """Set values for a variable.
        set_value(name, value)
        set_value(name, mapfrom=self, nomap=None)

        Parameters
        ----------
        name : str
            Name of the destination values.
        """
        if len(args) == 1:
            return super(GridMapperMixIn, self).set_value(name, *args)

        mapfrom = kwds.pop("mapfrom", self)
        nomap = kwds.pop("nomap", None)
        try:
            value, source = mapfrom
        except TypeError:
            value, source = name, mapfrom

        if nomap is not None:
            orig = self.get_value(name)

        data = source.regrid(value, to=self, to_name=name, **kwds)

        if nomap is not None:
            data[nomap] = orig[nomap]

        super(GridMapperMixIn, self).set_value(name, data)

    def map_value(self, name, **kwds):
        """Map values from another grid.

        Parameters
        ----------
        name : str
            Name of values to map to.
        mapfrom : tuple or bmi_like, optional
            BMI object from which values are mapped from. This can also be
            a tuple of *(name, bmi)*, where *name* is the variable of the
            source grid and *bmi* is the bmi-like source. If not provided,
            use *self*.
        nomap : narray of bool, optional
            Values in the destination grid to not map.
        """
        mapfrom = kwds.pop("mapfrom", self)
        nomap = kwds.pop("nomap", None)
        try:
            value, source = mapfrom
        except TypeError:
            value, source = name, mapfrom

        if nomap is not None:
            orig = self.get_value(name)

        data = source.regrid(value, to=self, to_name=name, **kwds)

        if nomap is not None:
            data[nomap] = orig[nomap]

        self.set_value(name, data)



import numpy as np

from pymt.grids.assertions import is_rectilinear, is_structured, is_unstructured


def get_default_coordinate_units(n_dims):
    if n_dims <= 0 or n_dims > 3:
        raise ValueError("dimension must be between one and three")
    return ["-"] * n_dims


def get_default_coordinate_names(n_dims):
    if n_dims <= 0 or n_dims > 3:
        raise ValueError("dimension must be between one and three")
    return ["z", "y", "x"][-n_dims:]


def assert_arrays_are_equal_size(*args):
    first_size = args[0].size
    for arg in args[1:]:
        if arg.size != first_size:
            raise AssertionError("arrays are not the same length")


def args_as_numpy_arrays(*args):
    np_arrays = []
    for arg in args:
        if isinstance(arg, np.ndarray):
            np_array = arg.view()
        else:
            np_array = np.array(arg)
        np_array.shape = (np_array.size,)
        np_arrays.append(np_array)
    return tuple(np_arrays)


def coordinates_to_numpy_matrix(*args):
    args = args_as_numpy_arrays(*args)
    assert_arrays_are_equal_size(*args)

    coords = np.empty((len(args), len(args[0])), dtype=np.float)
    for (dim, arg) in enumerate(args):
        coords[dim][:] = arg.flatten()
    return coords


def non_singleton_axes(grid):
    try:
        shape = grid.get_shape()
    except AttributeError:
        return np.arange(grid.get_dim_count())
    else:
        (indices,) = np.where(shape > 1)
        return indices


def non_singleton_shape(grid):
    shape = grid.get_shape()
    (indices,) = np.where(shape > 1)
    return shape[indices]


def non_singleton_coordinate_names(grid):
    indices = non_singleton_axes(grid)
    return grid.get_coordinate_name(indices)


def non_singleton_dimension_names(grid):
    if is_structured(grid, strict=False):
        coordinate_names = non_singleton_coordinate_names(grid)
        # return np.array(['n' + name for name in coordinate_names])
        return coordinate_names
    else:
        return np.array(["n_node"])


def non_singleton_dimension_shape(grid):
    if is_rectilinear(grid, strict=False):
        shape = non_singleton_dimension_names(grid)
        return shape[:, np.newaxis]
    elif is_structured(grid, strict=True):
        shape = non_singleton_dimension_names(grid)
        return np.tile(shape, (len(shape), 1))
    elif is_unstructured(grid, strict=True):
        shape = np.array(["n_node"])
        return np.tile(shape, (grid.get_dim_count(), 1))


def _find_first(array, value):
    try:
        return np.where(array == value)[0][0]
    except IndexError:
        return len(array)


def connectivity_matrix_as_array(face_nodes, bad_val):
    nodes_per_face = np.empty(face_nodes.shape[0], dtype=np.int)
    for (face_id, face) in enumerate(face_nodes):
        nnodes = _find_first(face, bad_val)
        if nnodes > 0:
            nodes_per_face[face_id] = _find_first(face, bad_val)
        else:
            raise ValueError("face contains no nodes")
    offsets = np.cumsum(nodes_per_face)

    connectivity = np.empty(offsets[-1], dtype=int)
    offset = 0
    for (n_nodes, face) in zip(nodes_per_face, face_nodes):
        connectivity[offset : offset + n_nodes] = face[:n_nodes]
        offset += n_nodes

    return (connectivity, offsets)



#! /bin/env python
"""
Examples
========

**Rectilinear**

Create a rectilinear grid that is 2x3::

    (0) --- (1) --- (2)
     |       |       |
     |       |       |
     |  [0]  |  [1]  |
     |       |       |
     |       |       |
    (3) --- (4) --- (5)

Numbers in parens are node IDs, and numbers in square brackets are
cell IDs.

>>> g = RectilinearMap ([0, 2], [0, 1, 2])
>>> g.get_x ()
array([ 0.,  1.,  2.,  0.,  1.,  2.])
>>> g.get_y ()
array([ 0.,  0.,  0.,  2.,  2.,  2.])

Node 1 is shared by both cell 0, and 1; node 5 only is part of cell 1.

>>> g.get_shared_cells (1)
[0, 1]
>>> g.get_shared_cells (5)
[1]

Point (.5, 1.) is contained only within cell 0.

>>> g.is_in_cell (.5, 1., 0)
True
>>> g.is_in_cell (.5, 1., 1)
False

Point (1., 1.) is on a border and so is contained by both cells.

>>> g.is_in_cell (1, 1., 0)
True
>>> g.is_in_cell (1, 1., 1)
True

"""

from shapely.geometry import Point, asLineString, asPoint, asPolygon

from pymt.grids import (
    Rectilinear,
    Structured,
    UniformRectilinear,
    Unstructured,
    UnstructuredPoints,
)


class UnstructuredMap(Unstructured):
    name = "Unstructured"

    def __init__(self, *args, **kwargs):
        super(UnstructuredMap, self).__init__(*args, **kwargs)

        self._point = {}
        last_offset = 0
        for (cell_id, offset) in enumerate(self._offset):
            cell = self._connectivity[last_offset:offset]
            last_offset = offset

            for point_id in cell:
                try:
                    self._point[point_id].append(cell_id)
                except KeyError:
                    self._point[point_id] = [cell_id]

        (point_x, point_y) = (self.get_x(), self.get_y())
        self._polys = []
        last_offset = 0
        for (cell_id, offset) in enumerate(self._offset):
            cell = self._connectivity[last_offset:offset]
            last_offset = offset

            (x, y) = (point_x.take(cell), point_y.take(cell))
            if len(x) > 2:
                self._polys.append(asPolygon(zip(x, y)))
            elif len(x) == 2:
                self._polys.append(asLineString(zip(x, y)))
            else:
                self._polys.append(asPoint(zip(x, y)))

    def get_shared_cells(self, point_id):
        """
        Parameters
        ----------
        point_id: int
            ID of a point in the grid.

        Returns
        -------
        ndarray of int
            Indices to cells that share a given node.
        """
        return self._point[point_id]

    def is_in_cell(self, x, y, cell_id):
        """Check if a point is in a cell.

        Parameters
        ----------
        x: float
            x-coordinate of point to check.
        y: float
            y-coordinate of point to check.
        cell_id: int
            ID of the cell in the grid.

        Returns
        -------
        bool
            True if the point (x, y) is contained in the cell.
        """
        pt = Point((x, y))
        return self._polys[cell_id].contains(pt) or self._polys[cell_id].touches(pt)


class UnstructuredPointsMap(UnstructuredPoints):
    name = "UnstructuredPoints"

    def get_shared_cells(self, point_id):  # pylint: disable=no-self-use
        return []

    def is_in_cell(self, x, y, cell_id):  # pylint: disable=no-self-use
        return False


class StructuredMap(Structured, UnstructuredMap):
    name = "Structured"


class RectilinearMap(Rectilinear, UnstructuredMap):
    name = "Rectilinear"


class UniformRectilinearMap(UniformRectilinear, UnstructuredMap):
    name = "UniformRectilinear"


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#! /bin/env python

from collections import defaultdict

import numpy as np
from scipy.spatial import KDTree
from six.moves import zip

from .imapper import IGridMapper, IncompatibleGridError

# from .mapper import IncompatibleGridError


def map_cells_to_points(coords, dst_grid, dst_point_ids, bad_val=-1):
    src_x, src_y = coords

    cell_to_point_id = defaultdict(list)
    for (j, point_id) in enumerate(dst_point_ids):
        for cell_id in dst_grid.get_shared_cells(point_id):
            if dst_grid.is_in_cell(src_x[j], src_y[j], cell_id):
                cell_to_point_id[cell_id].append(j)

    return cell_to_point_id


class PointToCell(IGridMapper):

    _name = "PointToCell"

    def initialize(self, dest_grid, src_grid, **kwds):
        if not self.test(dest_grid, src_grid):
            raise IncompatibleGridError(dest_grid.name, src_grid.name)
        src_x = src_grid.get_x()
        src_y = src_grid.get_y()

        tree = KDTree(list(zip(dest_grid.get_x(), dest_grid.get_y())))
        (_, nearest_dest_id) = tree.query(list(zip(src_x, src_y)))

        self._map = map_cells_to_points(
            (src_x, src_y), dest_grid, nearest_dest_id, bad_val=-1
        )

        self._dst_cell_count = dest_grid.get_cell_count()
        self._src_point_count = src_grid.get_point_count()

    def run(self, src_values, **kwds):
        dst_vals = kwds.get("dst_vals", None)
        bad_val = kwds.get("bad_val", -999)
        method = kwds.get("method", np.mean)

        if src_values.size != self._src_point_count:
            raise ValueError("size mismatch between source and point count")

        if dst_vals is None:
            dst_vals = np.array([bad_val] * self._dst_cell_count, dtype=np.float)
        if dst_vals.size != self._dst_cell_count:
            raise ValueError("size mismatch between destination and cell count")

        for (cell_id, point_ids) in self._map.items():
            if all(src_values[point_ids] > bad_val):
                dst_vals[cell_id] = method(src_values[point_ids])

        return dst_vals

    @staticmethod
    def test(dst_grid, src_grid):
        return all(np.diff(dst_grid.get_offset()) > 2) and src_grid is not None

    @property
    def name(self):
        return self._name



#! /usr/bin/env python
from six.moves import xrange

from ...grids import RectilinearField, StructuredField, UnstructuredField
from ...grids import utils as gutils
from .constants import open_netcdf


class NetcdfFieldReader(object):
    def __init__(self, path, fmt="NETCDF4"):
        self._path = path

        self._root = open_netcdf(path, mode="r", fmt=fmt)
        self._topology = self._get_mesh_topology()
        self._field = None
        self._time = []

        self._get_mesh_coordinate_data()
        self._get_node_variable_data()
        self._get_face_variable_data()
        if self.contains_time_dimension():
            self._get_time_variable()

        self._root.close()

    @property
    def fields(self):
        return self._field

    @property
    def times(self):
        return self._time

    def _get_mesh_coordinate_data(self):
        raise NotImplementedError("_get_mesh_coordinate_data")

    def contains_time_dimension(self):
        return "time" in self._root.dimensions

    def is_variable_data(self, name):
        var = self._root.variables[name]
        return (
            hasattr(var, "standard_name")
            and hasattr(var, "units")
            and hasattr(var, "coordinates")
            and hasattr(var, "location")
        )

    def variable_data_names(self):
        names = []
        for name in self._root.variables:
            if self.is_variable_data(name):
                names.append(name)
        return names

    def variable_data(self, name):
        return self._root.variables[name][:]

    def _get_mesh_topology(self):
        topology = self._root.variables["mesh"]

        if hasattr(topology, "type") and topology.type not in (
            None,
            "rectilinear",
            "structured",
            "unstructured",
        ):
            raise ValueError("topology not understood: {0}".format(topology.type))

        return topology

    def _get_node_variable_data(self):
        for name in self.variable_data_names():
            if self._root.variables[name].location == "node":
                data = self.variable_data(name)
                if self.contains_time_dimension():
                    for time in xrange(len(self.time)):
                        self._field.add_field(
                            name + "@t=%d" % time, data[time, :], centering="point"
                        )
                    self._field.add_field(name, data[-1, :], centering="point")
                else:
                    self._field.add_field(name, data[:], centering="point")

    def _get_face_variable_data(self):
        for name in self.variable_data_names():
            if self._root.variables[name].location == "face":
                data = self.variable_data(name)
                if self.contains_time_dimension():
                    for time in xrange(len(self.time)):
                        self._field.add_field(
                            name + "@t=%d" % time, data[time, :], centering="cell"
                        )
                    self._field.add_field(name, data[-1, :], centering="cell")
                else:
                    self._field.add_field(name, data[:], centering="cell")

    def _get_time_variable(self):
        if self.contains_time_dimension():
            self._time = self.variable_data("time")
        else:
            self._time = []

    @property
    def time(self):
        if self.contains_time_dimension():
            return self.variable_data("time")
        else:
            return []


class NetcdfRectilinearFieldReader(NetcdfFieldReader):
    def _get_mesh_coordinate_data(self):
        coordinate_names = self._topology.node_coordinates.split()
        coordinates = []
        for name in coordinate_names:
            coordinates.append(self._root.variables[name][:])
        self._field = RectilinearField(*coordinates)


class NetcdfStructuredFieldReader(NetcdfFieldReader):
    def _get_mesh_coordinate_data(self):
        coordinate_names = self._topology.node_coordinates.split()
        coordinates, shape = ([], [])
        for name in coordinate_names:
            coordinates.append(self._root.variables[name])
            shape.append(len(self._root.dimensions[name]))
        self._field = StructuredField(*(coordinates + [shape]))


class NetcdfUnstructuredFieldReader(NetcdfFieldReader):
    def face_nodes_data(self):
        name = self._topology.face_node_connectivity
        return self._root.variables[name]

    def face_nodes_fill_value(self):
        name = self._topology.face_node_connectivity
        return self._root.variables[name]._FillValue

    def face_nodes_start_index(self):
        name = self._topology.face_node_connectivity
        return self._root.variables[name].start_index

    def _get_mesh_coordinate_data(self):
        coordinate_names = self._topology.node_coordinates.split()
        coordinates = []
        for name in coordinate_names:
            coordinates.append(self._root.variables[name])

        (connectivity, offset) = gutils.connectivity_matrix_as_array(
            self.face_nodes_data(), self.face_nodes_fill_value
        )
        connectivity -= self.face_nodes_start_index()
        self._field = UnstructuredField(
            *coordinates, connectivity=connectivity, offset=offset
        )



import os
import string
import warnings

import numpy as np

from ..grids import RasterField, StructuredField, UnstructuredField


class Error(Exception):
    pass


class DimensionError(Error):
    def __init__(self, msg):
        self._msg = msg

    def __str__(self):
        return self._msg


class BadFileFormatError(Error):
    def __init__(self, fmt):
        self._format = fmt

    def __str__(self):
        return self._format


def positive_dimensions(shape):
    """Check if all dimensions are positive.

    Parameters
    ----------
    shape : ndarray of int
        Array shape.

    Returns
    -------
    boolean
        True if dimensions are positive non-zero. Otherwise, False.
    """
    return shape[shape > 0]


def find_unknown_dimension(shape):
    """Find unlimited dimensions.

    If a dimension is negative, it's thought to be *unknown* or *unlimited*
    Find that dimension. Raise an error if there are multiple unknown dimensions.

    Parameters
    ----------
    shape : ndarray of int
        Array shape.

    Returns
    -------
    int
        Dimension that is unknown or `None` if all dimensions are known.

    Raises
    ------
    DimensionError
        There are multiple unknown dimensions.
    """
    (unknown_dim,) = np.where(np.array(shape) < 0)
    if len(unknown_dim) > 1:
        raise DimensionError("more than one unknown dimension")
    try:
        return unknown_dim[0]
    except IndexError:
        return None


def fix_unknown_shape(shape, size):
    """Determine shape with an unknown dimension.

    Parameters
    ----------
    shape : ndarray of int
        Array shape (with a possible unknown dimension).
    size : int
        Sive of the array.

    Returns
    -------
    tuple of int
        Shape of the array with any unknown dimension fixed.

    Examples
    --------
    >>> from pymt.portprinter.utils import fix_unknown_shape

    >>> fix_unknown_shape((4, 3), 12)
    (4, 3)
    >>> fix_unknown_shape((4, -1), 12)
    (4, 3)
    """
    if len(shape) > 0:
        new_shape = np.array(shape)
    else:
        if size != 0:
            raise DimensionError("total size of new array must be unchanged")
        return (0,)

    known_element_count = np.prod(new_shape[new_shape > 0])
    if size % known_element_count != 0:
        raise DimensionError("total size of new array must be unchanged")

    unknown_dim = find_unknown_dimension(new_shape)
    if unknown_dim is not None:
        new_shape[unknown_dim] = size / known_element_count

    return tuple(new_shape)


def is_structured_port(port, grid_id):
    if hasattr(port, "get_grid_shape"):
        shape = port.get_grid_shape(grid_id)
        if shape is None:
            return False
    else:
        return False
    return True


def is_rectilinear_port(port, grid_id):
    if is_structured_port(port, grid_id) and hasattr(port, "get_grid_spacing"):
        spacing = port.get_grid_spacing(grid_id)
        if spacing is None:
            return False
    else:
        return False
    return True


def port_is_one_point(port, grid_id):
    shape = port.get_grid_shape(grid_id)
    if len(shape) == 1 and shape[0] == 1:
        return True
    else:
        return False


def _port_shape_as_array(port, grid_id):
    return port.get_grid_shape(grid_id)


def _port_spacing_as_array(port, grid_id):
    if port_is_one_point(port, grid_id):
        spacing = np.array([1.0])
    else:
        spacing = port.get_grid_spacing(grid_id)
    return spacing


def _port_origin_as_array(port, grid_id):
    if port_is_one_point(port, grid_id):
        origin = np.array([0.0])
    else:
        try:
            origin = port.get_grid_origin(grid_id)
        except AttributeError:
            origin = port.get_grid_lower_left(grid_id)
            warnings.warn("get_grid_lower_left", DeprecationWarning)
    return origin


def _construct_port_as_rectilinear_field(port, grid_id, data_array):
    if len(data_array) == 1:
        shape = np.array((1,))
        spacing = np.array((1.0,))
        origin = np.array((0.0,))
    else:
        shape = _port_shape_as_array(port, grid_id)
        spacing = _port_spacing_as_array(port, grid_id)
        origin = _port_origin_as_array(port, grid_id)

    shape = fix_unknown_shape(shape, data_array.size)

    return RasterField(shape, spacing, origin, indexing="ij")


def _construct_port_as_structured_field(port, grid_id, data_array):
    shape = _port_shape_as_array(port, grid_id)
    shape = fix_unknown_shape(shape, data_array.size)

    x = port.get_grid_x(grid_id)
    y = port.get_grid_y(grid_id)

    return StructuredField(x, y, shape)


def _construct_port_as_unstructured_field(port, grid_id):
    x = port.get_grid_x(grid_id)
    y = port.get_grid_y(grid_id)
    c = port.get_grid_connectivity(grid_id)
    o = port.get_grid_offset(grid_id)

    return UnstructuredField(x, y, c, o)


def construct_port_as_field(port, var_name):
    """Create a field object from a port.

    Parameters
    ----------
    port : port_like
        A port.
    var_name : str
        Name of the variable field.

    Returns
    -------
    field_like
        A newly created field that contains the data for *var_name*.
    """
    data_array = port.get_value(var_name)
    if data_array is None:
        raise ValueError(var_name)

    grid_id = port.get_var_grid(var_name)
    if len(data_array) == 1 or is_rectilinear_port(port, grid_id):
        field = _construct_port_as_rectilinear_field(port, grid_id, data_array)
    elif is_structured_port(port, grid_id):
        field = _construct_port_as_structured_field(port, grid_id, data_array)
    else:
        field = _construct_port_as_unstructured_field(port, grid_id)

    field.add_field(
        var_name, data_array, centering=get_data_centering(field, data_array)
    )

    return field


def get_data_centering(field, data_array):
    if data_is_centered_on_points(field, data_array):
        centering = "point"
    elif data_is_centered_on_cells(field, data_array):
        centering = "zonal"
    else:
        raise RuntimeError("statement should not be reached")
    return centering


def mesh_size_has_changed(field, data):
    return data.size not in [field.get_point_count(), field.get_cell_count()]


def data_is_centered_on_points(field, data):
    return data.size == field.get_point_count()


def data_is_centered_on_cells(field, data):
    return data.size == field.get_cell_count()


def reconstruct_port_as_field(port, field):
    """Recreate a field object from a port.

    Add data from *port* to *field*. If the mesh of the port is no longer the
    same as that of *field*, create a new field object and add the data to
    it. Where possible, the new data is added to the existing field. Only the
    data that does not fit on the existing field is added to the new one.

    Parameters
    ----------
    port : port_like
        A port.
    field : field_like
        An existing field with new data.

    Returns
    -------
    field_like
        A (possibley) newly created field that contains the data for *var_name*.
    """
    for var_name in field.keys():
        data_array = port.get_value(var_name)

        if mesh_size_has_changed(field, data_array):
            field = construct_port_as_field(port, var_name)
        else:
            field.add_field(
                var_name, data_array, centering=get_data_centering(field, data_array)
            )

    return field


_FORMAT_EXTENSION = {"nc": ".nc", "netcdf": ".nc"}


def normalize_format_name(fmt):
    if fmt.startswith("."):
        fmt = fmt[1:]
    return fmt.lower()


def format_to_file_extension(fmt):
    try:
        return _FORMAT_EXTENSION[normalize_format_name(fmt)]
    except KeyError:
        raise BadFileFormatError(normalize_format_name(fmt))


def _file_name_lacks_extension(file_name):
    (_, ext) = os.path.splitext(file_name)
    return len(ext) <= 1


def construct_file_name(var_name, template="${var}", fmt=None, prefix=""):
    file_template = string.Template(template)
    file_name = file_template.safe_substitute(var=var_name)

    if _file_name_lacks_extension(file_name):
        file_ext = format_to_file_extension(fmt)
        file_name = file_name + file_ext

    return os.path.join(prefix, file_name)


def next_unique_file_name(file_name):
    (base, ext) = os.path.splitext(file_name)
    count = 0
    while os.path.isfile(file_name):
        file_name = base + ".%d" % count + ext
        count += 1
    return file_name



import pkg_resources
import traceback

from .framework.bmi_bridge import bmi_factory


class ModelLoadError(Exception):
    def __init__(self, name, reason=None):
        self._name = name
        self._reason = reason or "no reason given"

    def __str__(self):
        return f"unable to load model ({self._name}):\n{self._reason}"


class ModelCollection:
    def __new__(cls):
        models = []
        errors = []
        for entry_point in pkg_resources.iter_entry_points(group="pymt.plugins"):
            try:
                models.append(ModelCollection.load_entry_point(entry_point))
            except ModelLoadError as error:
                errors.append((entry_point.name, str(error)))
        for name, model in models:
            setattr(cls, name, property(lambda self, name=name: self._data[name]))

        inst = super().__new__(cls)
        inst._errors = tuple(errors)
        inst._data = dict(models)

        return inst

    def __getitem__(self, name):
        return self._data[name]

    def keys(self):
        return self._data.keys()

    def items(self):
        return self._data.items()

    def values(self):
        return self._data.values()

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(set(self.keys()))

    @property
    def errors(self):
        return self._errors

    @staticmethod
    def load_entry_point(entry_point):
        try:
            model = entry_point.load()
        except Exception:
            raise ModelLoadError(entry_point.name, reason=traceback.format_exc())
        else:
            Model = bmi_factory(model)
            Model.__name__ = entry_point.name
            Model.__qualname__ = entry_point.name
            Model.__module__ = "pymt.models"

        return entry_point.name, Model
